import torch
import numpy as np
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from ase import Atoms, io
from ase.neighborlist import NeighborList
from typing import List, Optional, Union
from dataclasses import dataclass
from abfml.data.dataset import ABFMLDataset


@dataclass
class DataClass:
    def __init__(self, filename: str, file_format: Optional[str]):
        self.filename: str = filename
        if file_format == "pwmat-config":
            image_list = []
        elif file_format == "pwmat-movement":
            image_list = []
        else:
            image_list = io.read(filename, format=file_format, index=':')
        self.image_list: List[Atoms] = image_list
        self.include_element: set = set()
        include_atoms_number = []
        for atom_information in image_list:
            self.include_element = self.include_element.union(set(atom_information.get_chemical_symbols()))
            include_atoms_number.append(len(atom_information))
        self.include_atoms_number: set = set(include_atoms_number)
        self.n_frames: int = len(self.image_list)


class ReadData:
    def __init__(self, filename: Union[str, List[str]],  cutoff: float, neighbor: List[int],
                 type_map: Optional[List[int]], file_format: Optional[str]):
        self.cutoff = cutoff
        self.neighbor = neighbor
        self.filename = filename
        self.type_map = type_map
        self.image_list: list[Atoms] = []
        if isinstance(filename, str):
            filename = [filename]
        self.data_information: List[dict] = []
        for file in filename:
            data_information = DataClass(file, file_format)
            self.data_information.append({'file_name': data_information.filename,
                                          'n_frames': data_information.n_frames,
                                          'include_element': data_information.include_element,
                                          'include_atoms_number': data_information.include_atoms_number})
            self.image_list = self.image_list + data_information.image_list
        self.n_frames: int = len(self.image_list)
        self.unique_image = {}
        for i, image in enumerate(self.image_list):
            image_tuple = tuple(image.get_atomic_numbers())
            if image_tuple in self.unique_image:
                self.unique_image[image_tuple].append(i)
            else:
                self.unique_image[image_tuple] = [i]

    def get_mlffdata(self):
        dataset = []
        for same_image_index in self.unique_image.values():
            n_frames: int = len(same_image_index)
            n_atoms: int = len(self.image_list[same_image_index[0]])
            n_elements: int = len(set(self.image_list[same_image_index[0]].get_atomic_numbers()))
            if isinstance(self.neighbor, int):
                max_neighbor: int = self.neighbor
            elif isinstance(self.neighbor, list):
                if len(self.neighbor) == len(self.type_map):
                    max_neighbor = sum(self.neighbor[self.type_map.index(atomic_numbers)]
                                       for atomic_numbers in set(self.image_list[same_image_index[0]].get_atomic_numbers()))

                else:
                    raise Exception('neighbor and type_map must have the same length')
            else:
                raise ValueError(f'Expected a member of List[int] and int but instead found type {type(self.neighbor)}')

            Zi = torch.empty(n_frames, n_atoms, dtype=torch.int)
            Nij = torch.empty(n_frames, n_atoms, max_neighbor, dtype=torch.int)
            Zij = torch.empty(n_frames, n_atoms, max_neighbor, dtype=torch.int)
            Rij = torch.zeros(n_frames, n_atoms, max_neighbor, 4, dtype=torch.float64)
            element_type = torch.zeros(n_frames, n_elements, dtype=torch.int)
            energy = torch.zeros(n_frames, 1, dtype=torch.float64)
            atomic_energy = torch.zeros(n_frames, n_atoms, 1, dtype=torch.float64)
            force = torch.zeros(n_frames, n_atoms, 3, dtype=torch.float64)
            virial = torch.zeros(n_frames, 9, dtype=torch.float64)
            for i, image_index in enumerate(same_image_index):
                atom_ase = self.image_list[image_index]
                neighbor_information = ReadData.calculate_neighbor(atom=atom_ase,
                                                                   cutoff=self.cutoff,
                                                                   neighbor=self.neighbor,
                                                                   type_map=self.type_map)
                element_type = torch.tensor(neighbor_information[0])
                Zi[i] = torch.tensor(neighbor_information[1])
                Nij[i] = torch.tensor(neighbor_information[2])
                Zij[i] = torch.tensor(neighbor_information[3])
                Rij[i] = torch.tensor(neighbor_information[4])
                try:
                    energy[i] = torch.tensor(atom_ase.get_potential_energy())
                except (RuntimeError, AttributeError):
                    energy = None
                try:
                    atomic_energy[i] = torch.tensor(atom_ase.atomic_energy)
                except (RuntimeError, AttributeError):
                    atomic_energy = None
                try:
                    force[i] = torch.tensor(atom_ase.get_forces(apply_constraint=False))
                except (RuntimeError, AttributeError):
                    force = None
                try:
                    # ASE calculated in units of eV/A^3, virial: eV
                    virial[i] = torch.tensor(-1.0 * atom_ase.get_stress(voigt=False).reshape(9) * atom_ase.get_volume())
                except (RuntimeError, AttributeError):
                    virial = None
            dataset.append(ABFMLDataset(n_frames=n_frames,
                                        n_atoms=n_atoms,
                                        element_type=element_type,
                                        Zi=Zi,
                                        Nij=Nij,
                                        Zij=Zij,
                                        Rij=Rij,
                                        energy=energy,
                                        atomic_energy=atomic_energy,
                                        force=force,
                                        virial=virial))
        return dataset

    @staticmethod
    def calculate_neighbor(atom: Atoms, cutoff: float, neighbor: Union[List[int], int], type_map: Optional[List[int]]):
        nl = NeighborList([cutoff / 2] * len(atom), skin=0, self_interaction=False, bothways=True, sorted=False)
        nl.update(atom)
        atoms_num = len(atom)
        Zi = atom.numbers
        element_type = np.sort(np.unique(Zi))
        if isinstance(neighbor, int) or type_map is None:
            max_neighbor: int = neighbor
            Rij = np.zeros(shape=(1, atoms_num, max_neighbor, 4), dtype=np.float64)
            Nij = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            Zij = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            for i in range(atoms_num):
                indices_neighbor, offsets = nl.get_neighbors(i)
                j_xyz = atom.positions[indices_neighbor] + np.dot(offsets, atom.get_cell())
                ij_xyz = j_xyz - atom.positions[i]
                rij = np.sqrt(np.sum(ij_xyz ** 2, axis=1))
                if indices_neighbor.shape[0] <= max_neighbor:
                    end_index = indices_neighbor.shape[0]
                else:
                    indices_rij = np.argsort(rij, axis=-1)
                    for j in range(3):
                        ij_xyz[..., j] = np.take_along_axis(ij_xyz[..., j], indices=indices_rij, axis=-1)
                    rij = np.take_along_axis(rij, indices=indices_rij, axis=-1)
                    indices_neighbor = np.take_along_axis(indices_neighbor, indices=indices_rij, axis=-1)
                    end_index = max_neighbor
                Rij[:, i, :end_index, 1:] = ij_xyz[:end_index, :]
                Rij[:, i, :end_index, 0] = rij[:end_index]
                Nij[:, i, :end_index] = indices_neighbor[:end_index]
            mask = (Nij != -1)
            Zij[mask] = Zi[Nij[mask]]

        elif len(type_map) == len(neighbor):
            atom_map = type_map
            neighbor_map = neighbor
            Zi = atom.numbers
            max_neighbor = sum(neighbor_map[atom_map.index(atomic_numbers)] for atomic_numbers in element_type)
            width_map = [0]
            Zi_unique = list(element_type)
            for ii in Zi_unique:
                width_map.append(neighbor_map[atom_map.index(ii)]+width_map[-1])
            Rij = np.zeros(shape=(1, atoms_num, max_neighbor, 4), dtype=np.float64)
            Nij = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            Zij = np.full(shape=(1, atoms_num, max_neighbor), fill_value=-1, dtype=np.int32)
            for i in range(atoms_num):
                indices_neighbor, offsets = nl.get_neighbors(i)
                j_xyz = atom.positions[indices_neighbor] + np.dot(offsets, atom.get_cell())
                ij_xyz = j_xyz - atom.positions[i]
                rij = np.sqrt(np.sum(ij_xyz ** 2, axis=1))
                Zij_i = Zi[indices_neighbor]

                for atomic_numbers in Zi_unique:
                    mask = (Zij_i == atomic_numbers)
                    width = Zij_i[mask].shape[0]
                    start_index = width_map[Zi_unique.index(atomic_numbers)]
                    width_i = neighbor_map[Zi_unique.index(atomic_numbers)]
                    if width <= width_i:
                        end_index = start_index + width
                    else:
                        indices_rij = np.argsort(rij[mask], axis=-1)
                        for j in range(3):
                            ij_xyz[..., j][mask] = np.take_along_axis(ij_xyz[..., j][mask],
                                                                      indices=indices_rij,
                                                                      axis=-1)
                        rij[mask] = np.take_along_axis(rij[mask], indices=indices_rij, axis=-1)
                        indices_neighbor[mask] = np.take_along_axis(indices_neighbor[mask],
                                                                    indices=indices_rij,
                                                                    axis=-1)
                        end_index = start_index + width_i
                    Zij[:, i, start_index:end_index] = Zij_i[mask][:width_i]
                    Nij[:, i, start_index:end_index] = indices_neighbor[mask][:width_i]
                    Rij[:, i, start_index:end_index, 0] = rij[mask][:width_i]
                    Rij[:, i, start_index:end_index, 1:] = ij_xyz[mask][:width_i, :]
        else:
            raise Exception('neighbor[0] and neighbor[1] should have the same length '
                            'or neighbor[1] have only one element. Maybe you should read the manual!')
        Zi = Zi[np.newaxis,...]
        return element_type, Zi, Nij, Zij, Rij




import torch
import torch.nn as nn
import torch.optim as optim
from typing import Optional

# 定义简单神经网络
class SimpleNN():
    def __init__(self, input_size:Optional[int]=None, hidden_size:Optional[int]=None, output_size:Optional[int]=None):

        self.fc1 = 3 # 第一层
        self.relu = 1
        self.fc2 = 2
        self.input_size=input_size# 输出层

    def forward(self, x):
        x = self.fc1

        mapped_value = torch.arange(0, 10)
        a: int = int(torch.max(mapped_value).item())
        mapped_table = torch.full(size=( 1+ 1,2), fill_value=-1)
        if self.input_size is not None:
            x=x+1
        return x

torch.jit.script(SimpleNN(2, 3, 4))
print('aa')



from abfml.param.param import LossSet


def calculate_weight(param: LossSet, learn_rate: float):
    attributes = ['energy',  'ei', 'force', 'virial']
    weights = []

    for attr in attributes:
        start_weight = getattr(param, f'start_{attr}_weight')
        limit_weight = getattr(param, f'limit_{attr}_weight')
        weight = limit_weight * (1 - learn_rate) + start_weight * learn_rate
        weights.append(weight)

    return tuple(weights)






import time
import torch
import torch.nn as nn
from typing import List, Optional, Dict, Tuple, Any
from abfml.model.math_fun import smooth_fun, polynomial_fun, compute_forces_and_virial, calculate_distance_grad
from abfml.model.network import FittingNet
from abfml.model.method import FieldModel


@torch.jit.interface
class ModuleInterface(torch.nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass


class NEP(FieldModel):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int],
                 fitting_config: Dict[str, Any],
                 feature_config: Dict[str, Any],
                 energy_shift: List[float],
                 std_mean: Optional[List[torch.Tensor]],
                 ):
        super(NEP, self).__init__(type_map=type_map, cutoff=cutoff, neighbor=neighbor)
        self.std_mean = std_mean
        self.fitting_net = nn.ModuleList()
        self.fitting_net_index = []
        self.R_max_R = feature_config["R_max_R"]
        self.R_max_A = feature_config["R_max_A"]
        self.n_max_R = feature_config["n_max_R"]
        self.n_max_A = feature_config["n_max_A"]
        self.k_basis_R = feature_config["k_basis_R"]
        self.k_basis_A = feature_config["k_basis_A"]
        self.l_3b_max = feature_config["l_3b_max"]
        self.l_4b_max = feature_config["l_4b_max"]
        self.l_5b_max = feature_config["l_5b_max"]
        c2ij_nk = torch.zeros(max(type_map) + 1, max(type_map) + 1, self.n_max_R, self.k_basis_R + 1)
        c3ij_nk = torch.zeros(max(type_map) + 1, max(type_map) + 1, self.n_max_A, self.k_basis_A + 1)
        feature_num = (self.n_max_R + self.n_max_A * self.l_3b_max) * len(type_map)
        for i, element in enumerate(self.type_map):
            self.fitting_net.append(FittingNet(network_size=[feature_num] + fitting_config["network_size"],
                                               activate=fitting_config["activate_function"],
                                               bias=fitting_config["bias"],
                                               resnet_dt=fitting_config["resnet_dt"],
                                               energy_shift=energy_shift[i]))
            self.fitting_net_index.append(str(element))
            c2ij_nk[element] = torch.rand(c2ij_nk[element].shape)
            c3ij_nk[element] = torch.rand(c3ij_nk[element].shape)
        self.c2ij_nk = nn.Parameter(c2ij_nk)
        self.c3ij_nk = nn.Parameter(c3ij_nk)

    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              Nij: torch.Tensor,
              Zij: torch.Tensor,
              image_dR: torch.Tensor,
              n_ghost: int):

        batch, n_atoms, max_neighbor = Nij.shape
        device = image_dR.device
        dtype = image_dR.dtype
        type_map_temp: torch.Tensor = element_map.to(torch.int64)
        type_map: List[int] = type_map_temp.tolist()

        feature = NEP.calculate_feature(type_map=self.type_map,
                                        c2ij_nk=self.c2ij_nk,
                                        c3ij_nk=self.c3ij_nk,
                                        R_max_R=self.R_max_R,
                                        R_max_A=self.R_max_A,
                                        n_max_R=self.n_max_R,
                                        n_max_A=self.n_max_A,
                                        k_basis_R=self.k_basis_R,
                                        k_basis_A=self.k_basis_A,
                                        l_3b_max=self.l_3b_max,
                                        l_4b_max=self.l_4b_max,
                                        l_5b_max=self.l_5b_max,
                                        element_map=element_map,
                                        Zi=Zi, Zij=Zij, Ri=image_dR)
        Ei = torch.zeros(batch, n_atoms, 1, dtype=dtype, device=device)
        for i, itype in enumerate(type_map):
            mask_itype = (Zi == itype)
            if not mask_itype.any():
                continue
            iifeat = feature[mask_itype].reshape(batch, -1, feature.shape[-1])
            ii = self.fitting_net_index.index(str(itype))
            fitting_net: ModuleInterface = self.fitting_net[ii]
            Ei_itype = fitting_net.forward(iifeat)
            Ei[mask_itype] = Ei_itype.reshape(-1, 1)
        Etot = torch.sum(Ei, dim=1)

        return Etot, Ei

    @staticmethod
    def calculate_fk_rij(R_max: float,
                         k_basis: int,
                         rij: torch.Tensor) -> torch.Tensor:
        batch, n_atoms, _, _ = rij.shape
        R_min = 0.0
        fc_rij = smooth_fun(smooth_type='cos', rij=rij, r_inner=R_min, r_outer=R_max)
        Tk = polynomial_fun(fun_name='chebyshev', n=k_basis, rij=rij, r_inner=R_min, r_outer=R_max)
        fk_rij = 0.5 * (Tk + 1) * fc_rij

        return fk_rij

    @staticmethod
    def calculate_blm(xij: torch.Tensor,
                      yij: torch.Tensor,
                      zij: torch.Tensor,
                      rij: torch.Tensor,
                      R_max: float,
                      l_3b_max: int) -> torch.Tensor:
        batch, n_atoms, max_neighbor, _ = rij.shape
        device = rij.device
        dtype = rij.dtype

        xij = xij.squeeze(-1)
        yij = yij.squeeze(-1)
        zij = zij.squeeze(-1)
        xij_2 = xij ** 2
        yij_2 = yij ** 2
        zij_2 = zij ** 2
        rij_2 = xij_2 + yij_2 + zij_2

        blm = torch.zeros(batch, n_atoms, max_neighbor, 24, dtype=dtype, device=device)
        blm[..., 0] = zij
        blm[..., 1] = xij
        blm[..., 2] = yij
        blm[..., 3] = 3 * zij_2 - rij_2
        blm[..., 4] = xij * zij
        blm[..., 5] = yij * zij
        blm[..., 6] = xij_2 - yij_2
        blm[..., 7] = 2 * xij * yij
        blm[..., 8] = (5 * zij_2 - 3 * rij_2) * zij
        blm[..., 9] = (5 * zij_2 - rij_2) * xij
        blm[..., 10] = (5 * zij_2 - rij_2) * yij
        blm[..., 11] = (xij_2 - yij_2) * zij
        blm[..., 12] = 2 * xij * yij * zij
        blm[..., 13] = (xij_2 - 3 * yij_2) * xij
        blm[..., 14] = (3 * xij_2 - yij_2) * yij
        blm[..., 15] = (35 * zij_2 - 30 * rij_2) * zij_2 + 3 * rij_2 ** 2
        blm[..., 16] = (7 * zij_2 - 3 * rij_2) * xij * zij
        blm[..., 17] = (7 * zij_2 - 3 * rij_2) * yij * zij
        blm[..., 18] = (7 * zij_2 - rij_2) * (xij_2 - yij_2)
        blm[..., 19] = (7 * zij_2 - rij_2) * 2 * xij * yij
        blm[..., 20] = (xij_2 - 3 * yij_2) * xij * zij
        blm[..., 21] = (3 * xij_2) * yij * zij
        blm[..., 22] = (xij_2 - yij_2) ** 2 - 4 * xij_2 * yij_2
        blm[..., 23] = 4 * (xij_2 - yij_2) * xij * yij

        mask = rij.squeeze(-1) > 1e-5
        blm[mask] = blm[mask] / rij[mask]

        return blm

    @staticmethod
    def calculate_s_nlm(blm: torch.Tensor,
                        gn_rij: torch.Tensor,
                        rij: torch.Tensor):

        batch, n_atoms, max_neighbor, _ = rij.shape
        device = rij.device
        dtype = rij.dtype
        mask_rij = (rij[:, :, :, 0:1] > 1e-5)
        rr = torch.zeros(rij[:, :, :, 0:1].shape, dtype=dtype, device=device)
        rr[mask_rij] = 1 / rij[:, :, :, 0:1][mask_rij]
        S_nlm = (gn_rij.unsqueeze(-1) * (blm * rr).unsqueeze(-2))

        return S_nlm

    @staticmethod
    def calculate_feature(type_map: List[int],
                          c2ij_nk: torch.Tensor,
                          c3ij_nk: torch.Tensor,
                          R_max_R: float,
                          R_max_A: float,
                          n_max_R: int,
                          n_max_A: int,
                          k_basis_R: int,
                          k_basis_A: int,
                          l_3b_max: int,
                          l_4b_max: int,
                          l_5b_max: int,
                          element_map: torch.Tensor,
                          Zi: torch.Tensor,
                          Zij: torch.Tensor,
                          Ri: torch.Tensor):
        batch, n_atoms, max_neighbor, _ = Ri.shape
        dtype = Ri.dtype
        device = Ri.device
        include_element_list = element_map.to(torch.int64)

        C3lm = torch.tensor(data=[0.238732414637843, 0.238732414637843, 0.238732414637843, 0.099471839432435,
                                  1.193662073189215, 1.193662073189215, 0.298415518297303, 0.298415518297303,
                                  0.139260575205408, 0.208890862808112, 0.208890862808112, 2.088908628081126,
                                  2.088908628081126, 0.348151438013521, 0.348151438013521, 0.011190581936148,
                                  0.447623277445955, 0.447623277445955, 0.223811638722977, 0.223811638722977,
                                  3.133362942121689, 3.133362942121689, 0.391670367765211, 0.391670367765211],
                            dtype=dtype, device=device)
        C4lm = torch.tensor(data=[-0.00749948082666, -0.13499065487995, 0.067495327439977, 0.404971964639861,
                                  -0.80994392927972],
                            dtype=dtype, device=device)

        C5lm = torch.tensor(data=[0.026596810706114, 0.053193621412227, 0.026596810706114], dtype=dtype, device=device)
        fk_rij = NEP.calculate_fk_rij(R_max=max(R_max_R, R_max_A),
                                      k_basis=max(k_basis_R, k_basis_A),
                                      rij=Ri[..., 0:1])
        fk_rij_2 = torch.zeros(fk_rij.shape, dtype=dtype, device=device)
        mask = Ri[..., 0] < R_max_R
        fk_rij_2[mask] = fk_rij[mask]
        c2ij_n = c2ij_nk[Zi.unsqueeze(-1), Zij]
        gn2_rij = (torch.mul(fk_rij.unsqueeze(-2), c2ij_n).squeeze(-1)).sum(dim=-1)
        gn = torch.zeros(batch, n_atoms, len(type_map) * n_max_R, dtype=dtype, device=device)
        for i, itype in enumerate(include_element_list):
            indices = type_map.index(itype)
            mask_gn = (Zij == itype)
            gn[:, :, n_max_R * indices:n_max_R * (indices + 1)] = torch.sum(gn2_rij * mask_gn.unsqueeze(-1), dim=2)
        feature = gn
        if l_3b_max > 0:
            blm = NEP.calculate_blm(xij=Ri[..., 1:2], yij=Ri[..., 2:3], zij=Ri[..., 3:4], rij=Ri[..., 0:1], R_max=6.0, l_3b_max=4)

            fk_rij_3 = torch.zeros(fk_rij.shape, dtype=dtype, device=device)
            mask = Ri[..., 0] < R_max_A
            fk_rij_3[mask] = fk_rij[mask]
            c3ij_n = c3ij_nk[Zi.unsqueeze(-1), Zij]
            gn3_rij = (torch.mul(fk_rij.unsqueeze(-2), c3ij_n).squeeze(-1)).sum(dim=-1)
            S_nlm = NEP.calculate_s_nlm(blm=blm,
                                        gn_rij=gn3_rij,
                                        rij=Ri[..., 0:1])
            Cs = torch.zeros(batch, n_atoms, len(type_map), n_max_A, 24, dtype=dtype, device=device)
            for i, itype in enumerate(include_element_list):
                indices = type_map.index(itype)
                mask_s = (Zij == itype)[..., None, None]
                Cs[:, :, indices] = torch.sum(S_nlm * mask_s, dim=2)

            Cs = Cs ** 2 * C3lm
            qnl = torch.zeros(batch, n_atoms, len(type_map), n_max_A, l_3b_max, dtype=dtype, device=device)
            qnl[..., 0] = Cs[..., 0:3].sum(dim=-1)
            qnl[..., 1] = Cs[..., 3:8].sum(dim=-1)
            qnl[..., 2] = Cs[..., 8:15].sum(dim=-1)
            qnl[..., 3] = Cs[..., 15:24].sum(dim=-1)
            qnl = qnl.reshape(batch, n_atoms, n_max_A * l_3b_max * len(type_map))
            feature = torch.cat([gn, qnl], dim=-1)
        return feature






import torch
from torch.utils.data import Dataset


class ABFMLDataset(Dataset):
    def __init__(self,
                 n_frames: int,
                 n_atoms: int,
                 element_type: torch.Tensor,
                 Zi: torch.Tensor,
                 Nij: torch.Tensor,
                 Zij: torch.Tensor,
                 Rij: torch.Tensor,
                 energy: torch.Tensor = None,
                 atomic_energy: torch.Tensor = None,
                 force: torch.Tensor = None,
                 virial: torch.Tensor = None,
                 ):
        super(ABFMLDataset, self).__init__()
        self.n_frames = n_frames
        self.n_atoms = n_atoms

        physics_name = ["energy", "atomic_energy", "force", "virial"]
        physics_var = [energy, atomic_energy, force, virial]
        physics_shape = [(n_frames, 1), (n_frames, n_atoms, 1), (n_frames, n_atoms, 3), (n_frames, 9)]
        for tensor_p, shape_p, name_p in zip(physics_var, physics_shape, physics_name):
            if tensor_p is not None:
                if not isinstance(tensor_p, torch.Tensor):
                    raise TypeError(f"If {name_p} is required, then the type of a should be torch.Tensor, "
                                    f"but its type is {type(tensor_p)}")
                if tensor_p.shape != shape_p:
                    raise ValueError(f"Expected shape of {name_p} is {shape_p},"
                                     f"but the shape you get is {tensor_p.shape}")
        self.atomic_energy = atomic_energy
        self.energy = energy
        self.force = force
        self.virial = virial

        neighbor_name = ["Zi", "Nij", "Zij", "Rij"]
        neighbor_var = [Zi, Nij, Zij, Rij]
        neighbor_shape = [(n_frames, n_atoms), (n_frames, n_atoms, Nij.shape[-1]),
                          (n_frames, n_atoms, Nij.shape[-1]), (n_frames, n_atoms, Nij.shape[-1], 4)]
        for tensor_n, shape_n, name_n in zip(neighbor_var, neighbor_shape, neighbor_name):
            if not isinstance(tensor_n, torch.Tensor):
                raise TypeError(f"The type of {name_n} should to be torch.Tensor, but its {type(tensor_n)}")
            if tensor_n.shape != shape_n:
                raise ValueError(f"Expected shape of {name_n} is {shape_n}, but the shape you get is {tensor_n.shape}")
        self.element_type = element_type
        self.Zi = Zi
        self.Nij = Nij
        self.Zij = Zij
        self.Rij = Rij

    def __len__(self):
        return self.n_frames

    def __getitem__(self, index):
        mlff_dict = {
            'n_atoms': self.n_atoms,
            'element_type': self.element_type,
            'Zi': self.Zi[index],
            'Nij': self.Nij[index],
            'Zij': self.Zij[index],
            'Rij': self.Rij[index],
        }
        optional_name = ["energy", "atomic_energy", "force", "virial"]
        optional_val = [self.energy, self.atomic_energy, self.force, self.virial]
        for tensor, name in zip(optional_val, optional_name):
            if tensor is not None:
                mlff_dict[name] = tensor[index]
        return mlff_dict



import logging
from abfml.data.read_data import ReadData


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name: str, fmt: str = ".4e", summary_type: str = ""):
        self.name = name
        self.fmt = fmt
        summary_type_list = ["NONE", "AVERAGE", "SUM", "COUNT"]
        if summary_type in summary_type_list:
            self.summary_type = summary_type
        else:
            raise ValueError(f"Invalid summary type : {summary_type}, supported only {summary_type_list}")
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n: int = 1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = f"{self.name}: {self.val:{self.fmt}}"
        return fmtstr.format(**self.__dict__)

    def summary(self):
        fmtstr = ""
        if self.summary_type == "NONE":
            fmtstr = f"{self.name}: {self.val:{self.fmt}}"
        elif self.summary_type == "AVERAGE":
            fmtstr = f"{self.name}: {self.avg:{self.fmt}}"
        elif self.summary_type == "SUM":
            fmtstr = f"{self.name}: {self.sum:{self.fmt}}"
        elif self.summary_type == "COUNT":
            fmtstr = f"{self.name}: {self.count:{self.fmt}}"
        else:
            raise ValueError("Invalid summary type %r" % self.summary_type)
        return fmtstr


# Set to singleton mode, only one logger is allowed to exist for the same inputs
class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Logger(metaclass=Singleton):
    def __init__(self, log_file):
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        fh = logging.FileHandler(filename=log_file, mode="w", encoding="utf-8")
        fh.setLevel(logging.INFO)

        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        formatter = logging.Formatter(fmt='[%(levelname)s]- %(message)s', datefmt='%m-%d %H:%M:%S')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        self.logger.addHandler(fh)
        self.logger.addHandler(ch)


def log_logo(logger:logging.Logger):
    logger.info(r"################################################################################################")
    logger.info(r"#                         _       ______   ________  ____    ____  _____                       #")
    logger.info(r"#                        / \     |_   _ \ |_   __  ||_   \  /   _||_   _|                      #")
    logger.info(r"#                       / _ \      | |_) |  | |_ \_|  |   \/   |    | |                        #")
    logger.info(r"#                      / ___ \     |  __'.  |  _|     | |\  /| |    | |   _                    #")
    logger.info(r"#                    _/ /   \ \_  _| |__) |_| |_     _| |_\/_| |_  _| |__/ |                   #")
    logger.info(r"#                   |____| |____||_______/|_____|   |_____||_____||________|                   #")
    logger.info(r"#                                                                                              #")
    logger.info(r"################################################################################################")


def log_data_info(logger: logging.Logger, dataclass: ReadData):
    total_frames = 0
    logger.info("| ################ file name ############### Atom number ## nframes ######### Element ######## |")
    for data_information in dataclass.data_information:
        file_name = data_information['file_name'][-45:]
        n_frames = data_information['n_frames']
        chemical_symbols = ','.join(data_information['include_element']) if len(data_information['include_element']) \
                           <= 5 else 'complex'
        n_atoms = str(data_information['include_atoms_number'].pop()) if len(data_information['include_atoms_number']) == 1 \
            else 'variation'
        logger.info(f"|{file_name:>49s}{n_atoms:>6s}{n_frames:>8d}{chemical_symbols:>18s}")
        total_frames += n_frames
    logger.info(f"| Number of files found: {len(dataclass.data_information):>4d}, number of image : {total_frames:>4d}")
    logger.info(f"+----------------------------------------------------------------------------------------------+")




import time
import torch
import torch.nn as nn
from typing import List, Optional, Dict, Tuple, Any
from abfml.model.math_fun import smooth_fun
from abfml.model.network import FittingNet, AtomFitNet
from abfml.model.method import FieldModel


@torch.jit.interface
class ModuleInterface(torch.nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass


class BPMlp(FieldModel):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int],
                 fit_config: Dict[str, Any],
                 bp_features_information: List[Tuple[str, int]],
                 bp_features_param: List[Tuple[Dict[str, torch.Tensor], Dict[str, str]]],
                 energy_shift: List[float],
                 std_mean: Optional[List[torch.Tensor]],
                 ):
        super(BPMlp, self).__init__(type_map=type_map, cutoff=cutoff, neighbor=neighbor)
        self.type_map = type_map
        self.neighbor = neighbor
        self.cutoff = cutoff
        self.std_mean = std_mean
        self.bp_features_information = bp_features_information
        self.bp_features_param = bp_features_param
        self.fitting_net = nn.ModuleList()
        self.fitting_net_index = []

        total_feature_num = BPMlp.calculate_feature_num(num_element=len(type_map),
                                                        bp_features_information=bp_features_information)
        for i, element in enumerate(self.type_map):
            self.fitting_net.append(FittingNet(network_size=[total_feature_num] + fit_config["network_size"],
                                               activate=fit_config["activate_function"],
                                               bias=fit_config["bias"],
                                               resnet_dt=fit_config["resnet_dt"],
                                               energy_shift=energy_shift[i]))
            self.fitting_net_index.append(str(element))

    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              Nij: torch.Tensor,
              Zij: torch.Tensor,
              Rij: torch.Tensor,
              n_ghost: int):
        batch, n_atoms, max_neighbor = Nij.shape
        device = Rij.device
        dtype = Rij.dtype
        type_map_temp: torch.Tensor = element_map.to(torch.int64)
        type_map: List[int] = type_map_temp.tolist()
        feature = BPMlp.calculate_bp_feature(type_map=self.type_map,
                                             bp_features_information=self.bp_features_information,
                                             bp_features_param=self.bp_features_param,
                                             element_map=element_map,
                                             Nij=Nij,
                                             Zij=Zij,
                                             Rij=Rij)
        feature = BPMlp.scale(self.type_map, type_map, self.std_mean, feature, Zi)
        Ei = torch.zeros(batch, n_atoms, 1, dtype=dtype, device=device)
        for i, itype in enumerate(type_map):
            mask_itype = (Zi == itype)
            if not mask_itype.any():
                continue
            iifeat = feature[mask_itype].reshape(batch, -1, feature.shape[-1])
            ii = self.fitting_net_index.index(str(itype))
            fitting_net: ModuleInterface = self.fitting_net[ii]
            Ei_itype = fitting_net.forward(iifeat)
            Ei[mask_itype] = Ei_itype.reshape(-1, 1)
        Etot = torch.sum(Ei, dim=1)
        return Etot, Ei

    @staticmethod
    def scale(type_map_all: List[int],
              type_map_use: List[int],
              std_mean: List[torch.Tensor],
              feature: torch.Tensor,
              Zi: torch.Tensor) -> torch.Tensor:
        for i, element in enumerate(type_map_use):
            device = feature.device
            indices = type_map_all.index(element)
            mask = (Zi == element)
            std = std_mean[0][indices].detach().to(device)
            avg = std_mean[1][indices].detach().to(device)
            feature[mask] = (feature[mask] - avg) / std
        return feature

    @staticmethod
    def calculate_bp_feature(type_map: List[int],
                             bp_features_information: List[Tuple[str, int]],
                             bp_features_param: List[Tuple[Dict[str, torch.Tensor], Dict[str, str]]],
                             element_map: torch.Tensor,
                             Nij: torch.Tensor,
                             Zij: torch.Tensor,
                             Rij: torch.Tensor) -> torch.Tensor:
        feature_list: List[torch.Tensor] = []
        for i, feature_information in enumerate(bp_features_information):
            feature_name = feature_information[0]
            if feature_name == "SymFunRad":
                Rad_feature = BPMlp.sym_rad(type_map=type_map,
                                            R_min=(bp_features_param[i][0]['R_min']).item(),
                                            R_max=(bp_features_param[i][0]['R_max']).item(),
                                            eta=bp_features_param[i][0]['eta'],
                                            rs=bp_features_param[i][0]['rs'],
                                            smooth_type=bp_features_param[i][1]['smooth_fun'],
                                            element_map=element_map,
                                            Nij=Nij,
                                            Zij=Zij,
                                            Rij=Rij)
                feature_list.append(Rad_feature)

            if feature_name == "SymFunAngW":
                Ang_w_feature = BPMlp.sym_ang_w(type_map=type_map,
                                                R_min=(bp_features_param[i][0]['R_min'].item()),
                                                R_max=(bp_features_param[i][0]['R_max'].item()),
                                                lambd=bp_features_param[i][0]['lambd'],
                                                eta=bp_features_param[i][0]['eta'],
                                                zeta=bp_features_param[i][0]['zeta'],
                                                rs=bp_features_param[i][0]['rs'],
                                                smooth_type=bp_features_param[i][1]['smooth_fun'],
                                                element_map=element_map,
                                                Nij=Nij,
                                                Zij=Zij,
                                                Rij=Rij)
                feature_list.append(Ang_w_feature)

            if feature_name == "SymFunAngN":
                Ang_n_feature = BPMlp.sym_ang_n(type_map=type_map,
                                                R_min=(bp_features_param[i][0]['R_min'].item()),
                                                R_max=(bp_features_param[i][0]['R_max'].item()),
                                                lambd=bp_features_param[i][0]['lambd'],
                                                eta=bp_features_param[i][0]['eta'],
                                                zeta=bp_features_param[i][0]['zeta'],
                                                rs=bp_features_param[i][0]['rs'],
                                                smooth_type=bp_features_param[i][1]['smooth_fun'],
                                                element_map=element_map,
                                                Nij=Nij,
                                                Zij=Zij,
                                                Rij=Rij)
                feature_list.append(Ang_n_feature)

        feature = torch.cat([tensor for tensor in feature_list], dim=-1)
        return feature

    @staticmethod
    def calculate_feature_num(num_element: int, bp_features_information: List[Tuple[str, int]]) -> int:
        feature_num = 0
        for i, feature_information in enumerate(bp_features_information):
            feature_name = feature_information[0]
            num_basis = feature_information[1]
            if feature_name == "SymFunRad":
                feature_num += num_basis * num_element
            if "SymFunAng" in feature_name:
                feature_num += num_basis * num_element ** 2
        return feature_num

    @staticmethod
    def sym_rad(type_map: List[int],
                R_min: float,
                R_max: float,
                eta: torch.Tensor,
                rs: torch.Tensor,
                smooth_type: str,
                element_map: torch.Tensor,
                Nij: torch.Tensor,
                Zij: torch.Tensor,
                Rij: torch.Tensor) -> torch.Tensor:
        batch, n_atoms, max_neighbor = Nij.shape
        n_basis = eta.shape[0]
        device = Rij.device
        dtype = Rij.dtype
        include_element_list: List[int] = element_map.to(torch.int64).tolist()

        rij = Rij[:, :, :, 0].unsqueeze(-1)

        f_rij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        g_rij = torch.exp(-eta * (rij - rs) ** 2) * f_rij
        Gi = torch.zeros(batch, n_atoms, int(n_basis * len(type_map)), dtype=dtype, device=device)

        for i, itype in enumerate(include_element_list):
            indices = type_map.index(itype)
            mask_g = (Zij == itype).unsqueeze(-1)
            Gi[:, :,  n_basis * indices:n_basis * (indices + 1)] = torch.sum(g_rij * mask_g, dim=2)

        return Gi

    @staticmethod
    def sym_ang_n(type_map: List[int],
                  R_min: float,
                  R_max: float,
                  lambd: torch.Tensor,
                  eta: torch.Tensor,
                  zeta: torch.Tensor,
                  rs: torch.Tensor,
                  smooth_type: str,
                  element_map: torch.Tensor,
                  Nij: torch.Tensor,
                  Zij: torch.Tensor,
                  Rij: torch.Tensor) -> torch.Tensor:
        batch, n_atoms, max_neighbor = Nij.shape
        n_basis = eta.shape[0]
        device = Rij.device
        dtype = Rij.dtype
        include_element_list: List[int] = element_map.to(torch.int64).tolist()
        rij = Rij[:, :, :, 0:1]
        # Guaranteed not to divide by 0
        mask_rij = (rij > 1e-5)
        rr = torch.zeros(rij.shape, dtype=dtype, device=device)
        rr[mask_rij] = 1 / rij[mask_rij]

        frij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        grij = torch.exp(-eta * (rij - rs) ** 2) * frij
        cos_ijk = (torch.matmul(Rij[:, :, :, 1:], Rij[:, :, :, 1:].transpose(-1, -2))
                   * (rr * rr.transpose(-2, -1))).unsqueeze(2)
        Gi = torch.zeros(batch, n_atoms, int(n_basis * len(type_map) ** 2), dtype=dtype, device=device)

        for i, itype in enumerate(include_element_list):
            i_indices = type_map.index(itype)
            mask_i = (Zij == itype).unsqueeze(-1)
            gij = grij * mask_i
            # gij [batch, n_atoms, n_basis, max_neighbor, 1]
            gij = gij.transpose(2, 3).unsqueeze(-1)

            zeta = zeta.reshape(1, 1, n_basis, 1, 1)
            lambd = lambd.reshape(1, 1, n_basis, 1, 1)
            # t41 = time.time()
            for j, jtype in enumerate(include_element_list):
                j_indices = type_map.index(jtype)
                indices = i_indices * len(type_map) + j_indices
                mask_j = (Zij == itype).unsqueeze(-1)
                gik = grij * mask_j
                # gik [batch, n_atoms, n_basis, 1, max_neighbor]
                gik = gik.transpose(2, 3).unsqueeze(-2)
                # Gi_temp [batch, n_atoms, n_basis, max_neighbor, max_neighbor]
                Gi_temp = 2 ** (1 - zeta) * ((1 + lambd * cos_ijk) ** zeta) * (gij * gik)
                Gi[:, :,  n_basis * indices:n_basis * (indices + 1)] = Gi_temp.sum(dim=[-1, -2])
        return Gi

    @staticmethod
    def sym_ang_w(type_map: List[int],
                  R_min: float,
                  R_max: float,
                  lambd: torch.Tensor,
                  eta: torch.Tensor,
                  zeta: torch.Tensor,
                  rs: torch.Tensor,
                  smooth_type: str,
                  element_map: torch.Tensor,
                  Nij: torch.Tensor,
                  Zij: torch.Tensor,
                  Rij: torch.Tensor):
        batch, n_atoms, max_neighbor = Nij.shape
        n_basis = eta.shape[0]
        device = Rij.device
        dtype = Rij.dtype
        include_element_list: List[int] = element_map.to(torch.int64).tolist()
        rij = Rij[:, :, :, 0:1]

        # Guaranteed not to divide by 0
        mask_rij = (rij > 1e-5)
        rr = torch.zeros(rij.shape, dtype=dtype, device=device)
        rr[mask_rij] = 1 / rij[mask_rij]

        frij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        grij = torch.exp(-eta * (rij - rs) ** 2) * frij
        # cos_ijk [batch, n_atoms, 1,max_neighbor, max_neighbor]
        cos_ijk = (torch.matmul(Rij[:, :, :, 1:], Rij[:, :, :, 1:].transpose(-1, -2))
                   * (rr * rr.transpose(-2, -1))).unsqueeze(2)
        rjk2 = ((Rij[:, :, :, 1:].unsqueeze(3) - Rij[:, :, :, 1:].unsqueeze(3).permute(0, 1, 3, 2, 4)) ** 2).sum(-1)
        mask_rjk = (rjk2 > 1e-5)
        rjk = torch.zeros(rjk2.shape, dtype=dtype, device=device)
        rjk[mask_rjk] = rjk[mask_rjk] ** 0.5
        frjk = smooth_fun(smooth_type=smooth_type, rij=rjk, r_inner=R_min, r_outer=R_max)
        grjk = (torch.exp(-eta * (rjk.unsqueeze(-1) - rs) ** 2) * frjk.unsqueeze(-1)).transpose(2, 4).squeeze(-1)

        Gi = torch.zeros(batch, n_atoms, int(n_basis * len(type_map) ** 2), dtype=dtype, device=device)
        for i, itype in enumerate(include_element_list):
            i_indices = type_map.index(itype)
            mask_i = (Zij == itype).unsqueeze(-1)
            gij = grij * mask_i
            # gij [batch, n_atoms, n_basis, max_neighbor, 1]
            gij = gij.transpose(2, 3).unsqueeze(-1)

            zeta = zeta.reshape(1, 1, n_basis, 1, 1)
            lambd = lambd.reshape(1, 1, n_basis, 1, 1)

            # t41 = time.time()
            for j, jtype in enumerate(include_element_list):
                j_indices = type_map.index(jtype)
                indices = i_indices * len(type_map) + j_indices
                mask_j = (Zij == itype).unsqueeze(-1)
                gik = grij * mask_j
                # gik [batch, n_atoms, n_basis, 1, max_neighbor]
                gik = gik.transpose(2, 3).unsqueeze(-2)
                # Gi_temp [batch, n_atoms, n_basis, max_neighbor, max_neighbor]
                Gi_temp = 2 ** (1 - zeta) * ((1 + lambd * cos_ijk) ** zeta) * (gij * gik * grjk)
                Gi[:, :,  n_basis * indices:n_basis * (indices + 1)] = Gi_temp.sum(dim=[-1, -2])
        return Gi


class BPNN(FieldModel):
    """
    This is a reference to the “https://github.com/cc-ats/mlp_tutorial” tutorial.
    """
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int],
                 config: Dict[str, Any],
                 normal: Optional[List[torch.Tensor]],
                 ):
        super(BPNN, self).__init__(type_map=type_map, cutoff=cutoff, neighbor=neighbor)
        self.type_map = type_map
        self.neighbor = neighbor
        self.cutoff = cutoff
        descriptor_name = [descriptor_name['type_name'] for descriptor_name in config['list']]
        g1_config, g2_config, g3_config = None, None, None
        if 'SymFunRad' in descriptor_name:
            g1_config = next((config_dict for config_dict in config['list'] if config_dict.get('type_name') == 'SymFunRad'), None)
            g1_config['rs'] =  torch.Tensor(g1_config['rs'])
            g1_config['eta'] = torch.Tensor(g1_config['eta'])
            if g1_config['eta'].shape[0] != len(type_map):
                g1_config['rs'] = g1_config['rs'].repeat(len(type_map), 1)
                g1_config['eta'] = g1_config['eta'].repeat(len(type_map), 1)
        if 'SymFunAng' in descriptor_name:
            g2_config = next((config_dict for config_dict in config['list'] if config_dict.get('type_name') == 'SymFunAng'), None)
            g2_config['rs'] = torch.Tensor(g2_config['rs'])
            g2_config['eta'] = torch.Tensor(g2_config['eta'])
            g2_config['lambd'] = torch.Tensor(g2_config['lambd'])
            g2_config['zeta'] = torch.Tensor(g2_config['zeta'])
            if g2_config['zeta'].shape[0] != len(type_map):
                g2_config['lambd'] = g2_config['lambd'].repeat(len(type_map), 1)
                g2_config['zeta'] = g2_config['zeta'].repeat(len(type_map), 1)
                g2_config['rs'] = g2_config['rs'].repeat(len(type_map), 1)
                g2_config['eta'] = g2_config['eta'].repeat(len(type_map), 1)
        if 'SymFunAngN' in descriptor_name:
            g3_config = next((config_dict for config_dict in config['list'] if config_dict.get('type_name') == 'SymFunAngN'), None)
            g3_config['rs'] = torch.Tensor(g3_config['rs'])
            g3_config['eta'] = torch.Tensor(g3_config['eta'])
            g3_config['lambd'] = torch.Tensor(g3_config['lambd'])
            g3_config['zeta'] = torch.Tensor(g3_config['zeta'])
            if g3_config['zeta'].shape[0] != len(type_map):
                g3_config['lambd'] = g3_config['lambd'].repeat(len(type_map), 1)
                g3_config['zeta'] = g3_config['zeta'].repeat(len(type_map), 1)
                g3_config['rs'] = g3_config['rs'].repeat(len(type_map), 1)
                g3_config['eta'] = g3_config['eta'].repeat(len(type_map), 1)

        if descriptor_name == ['SymFunRad','SymFunAng','SymFunAngN']:
            self.descriptor = Feature()
        elif descriptor_name == ['SymFunRad','SymFunAngN']:
            self.descriptor = Feature(g1_Rmax=g1_config['R_max'], g1_Rmin=g1_config['R_min'], g1_eta=g1_config['eta'],
                                             g1_smooth_type=g1_config['smooth_type'], g1_rs=g1_config['rs'],
                                             g2_Rmax=g2_config['R_max'], g2_Rmin=g2_config['R_min'], g2_eta=g2_config['eta'],
                                             g2_smooth_type=g2_config['smooth_fun'], g2_lambd=g2_config['lambd'],
                                             g2_theta=g2_config['theta'], g2_rs=g2_config['rs'])
        elif descriptor_name == ['SymFunRad']:
            self.descriptor= Feature(g1_Rmax=g1_config['R_max'],g1_Rmin=g1_config['R_min'],g1_eta=g1_config['eta'],
                                             g1_smooth_type=g1_config['smooth_fun'],g1_rs=g1_config['rs'])
        else:
            raise Exception(f"Symmetric function configuration error")
        feature_length = self.descriptor.feature_length
        self.fitting_net = AtomFitNet(num_channels=len(type_map), network_size= [feature_length,200,220,100,1],
                                      activate='tanh',bias=True, resnet_dt= False)


    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              Nij: torch.Tensor,
              Zij: torch.Tensor,
              Rij: torch.Tensor,
              n_ghost: int):
        device = Rij.device
        dtype = Rij.dtype
        type_map_tensor = torch.tensor(self.type_map)
        mapped_value = torch.arange(0, len(self.type_map))
        mapped_table = torch.full(size=(int(torch.max(type_map_tensor).item()) + 1, ), fill_value=-1).to(device=device)
        mapped_table[type_map_tensor] = mapped_value
        mapped_Zi = mapped_table[Zi]
        feature = self.descriptor(mapped_Zi=mapped_Zi, Rij=Rij)
        Ei = self.fitting_net(feature, mapped_Zi)
        Etot = torch.sum(Ei,dim=1)
        return Etot, Ei


class Feature(nn.Module):
    def __init__(self,
                 g1_Rmin: Optional[float] = None,
                 g1_Rmax: Optional[float] = None,
                 g1_eta: Optional[torch.Tensor] = None,
                 g1_smooth_type: Optional[str] = None,
                 g1_rs: Optional[torch.Tensor] = None,
                 g2_Rmax: Optional[float] = None,
                 g2_Rmin: Optional[float] = None,
                 g2_eta: Optional[torch.Tensor] = None,
                 g2_smooth_type: Optional[str] = None,
                 g2_lambd: Optional[torch.Tensor] = None,
                 g2_theta: Optional[torch.Tensor] = None,
                 g2_rs: Optional[torch.Tensor] = None,
                 g3_Rmin: Optional[float] = None,
                 g3_Rmax: Optional[float] = None,
                 g3_eta: Optional[torch.Tensor] = None,
                 g3_smooth_type: Optional[str] = None,
                 g3_lambd: Optional[torch.Tensor] = None,
                 g3_theta: Optional[torch.Tensor] = None,
                 g3_rs: Optional[torch.Tensor] = None) -> None:
        """
        Initialize the Feature class with the specified parameters.

        Parameters for g1 (radial symmetry functions):
        - g1_Rmin, g1_Rmax: Radial distance range.
        - g1_eta: Scaling parameter for radial functions.
        - g1_smooth_type: Type of smoothing function ('cosine' or 'linear').
        - g1_rs: Shift parameters for radial symmetry functions.

        Parameters for g2 (angular symmetry functions with angular dependencies):
        - g2_Rmin, g2_Rmax: Distance range.
        - g2_eta: Scaling parameter for angular functions.
        - g2_smooth_type: Type of smoothing function ('cosine' or 'linear').
        - g2_lambd: Lambda parameter for angular symmetry.
        - g2_theta: Angular shift parameters.
        - g2_rs: Shift parameters for angular symmetry functions.

        Parameters for g3 (angular symmetry functions with triple dependencies):
        - g3_Rmin, g3_Rmax: Distance range.
        - g3_eta: Scaling parameter for triple angular functions.
        - g3_smooth_type: Type of smoothing function ('cosine' or 'linear').
        - g3_lambd: Lambda parameter for triple angular symmetry.
        - g3_theta: Angular shift parameters.
        - g3_rs: Shift parameters for triple angular symmetry functions.
        """
        super().__init__()
        # Radial symmetry function parameters
        self.g1_Rmin = g1_Rmin
        self.g1_Rmax = g1_Rmax
        self.g1_eta = g1_eta
        self.g1_smooth_type = g1_smooth_type
        self.g1_rs = g1_rs

        # Angular symmetry function parameters (pairwise angular dependencies)
        self.g2_Rmin = g2_Rmin
        self.g2_Rmax = g2_Rmax
        self.g2_eta = g2_eta
        self.g2_smooth_type = g2_smooth_type
        self.g2_lambd = g2_lambd
        self.g2_theta = g2_theta
        self.g2_rs = g2_rs

        # Angular symmetry function parameters (triple dependencies)
        self.g3_Rmin = g3_Rmin
        self.g3_Rmax = g3_Rmax
        self.g3_eta = g3_eta
        self.g3_smooth_type = g3_smooth_type
        self.g3_lambd = g3_lambd
        self.g3_theta = g3_theta
        self.g3_rs = g3_rs

    def forward(self, Rij: torch.Tensor, mapped_Zi: torch.Tensor) -> torch.Tensor:
        device, dtype = Rij.device, Rij.dtype
        feature_list = []
        # Compute g1 features if parameters are set
        if self.g1_eta is not None:
            g1_eta = self.g1_eta.to(device=device, dtype=dtype)[mapped_Zi]
            g1_rs = self.g1_rs.to(device=device, dtype=dtype)[mapped_Zi]
            feature_list.append(Feature.sym_rad(R_min=self.g1_Rmin, R_max=self.g1_Rmax, eta=g1_eta,
                                                smooth_type=self.g1_smooth_type, rs=g1_rs, Rij=Rij))

        # Prepare parameters for g2
        if self.g2_eta is not None:
            g2_eta = self.g2_eta.to(device=device, dtype=dtype)[mapped_Zi]
            g2_lambd = self.g2_lambd.to(device=device, dtype=dtype)[mapped_Zi]
            g2_theta = self.g2_theta.to(device=device, dtype=dtype)[mapped_Zi]
            g2_rs = self.g2_rs.to(device=device, dtype=dtype)[mapped_Zi]
            g2_zeta = self.g2_zeta.to(device=device, dtype=dtype)[mapped_Zi]

            feature_list.append(Feature.sym_ang_n(R_min=self.g2_Rmin, R_max=self.g2_Rmax, lambd=g2_lambd,eta=g2_eta,
                                                  zeta=g2_zeta, theta=g2_theta, rs=g2_rs,smooth_type=self.g2_smooth_type,
                                                  Rij=Rij))

        # Prepare parameters for g3
        if self.g3_eta is not None:
            g3_eta = self.g3_eta.to(device=device, dtype=dtype)[mapped_Zi]
            g3_lambd = self.g3_lambd.to(device=device, dtype=dtype)
            g3_theta = self.g3_theta.to(device=device, dtype=dtype)
            g3_rs = self.g3_rs.to(device=device, dtype=dtype)

            feature_list.append(Feature.sym_ang_w(R_min=self.g3_Rmin, R_max=self.g3_Rmax, lambd=g3_lambd, eta=g3_eta,
                                                  zeta=g3_theta,rs=g3_rs,smooth_type=self.g3_smooth_type,Rij=Rij))

        # Concatenate all computed features
        feature = torch.cat([tensor for tensor in feature_list], dim=-1)
        return feature

    @property
    def feature_length(self)->int:
        lengths = self.g1_eta.shape[-1]
        if self.g2_eta is not None:
            lengths += self.g2_eta[-1]
        if self.g3_eta is not None:
            lengths += self.g3_eta[-1]
        return lengths

    @staticmethod
    def sym_rad(R_min: float,
                R_max: float,
                eta: torch.Tensor,
                rs: torch.Tensor,
                smooth_type: str,
                Rij: torch.Tensor) -> torch.Tensor:
        rij = Rij[:, :, :, 0:1]
        f_rij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        g_rij = torch.exp(-eta.unsqueeze(dim=2) * (rij - rs.unsqueeze(dim=2)) ** 2) * f_rij
        Gi =  torch.sum(g_rij, dim=2)
        return Gi

    @staticmethod
    def sym_ang_n(R_min: float,
                  R_max: float,
                  lambd: torch.Tensor,
                  eta: torch.Tensor,
                  zeta: torch.Tensor,
                  theta: torch.Tensor,
                  rs: torch.Tensor,
                  smooth_type: str,
                  Rij: torch.Tensor) -> torch.Tensor:

        Rij = Rij[:, :, :, 1:4]
        c = torch.combinations(torch.arange(Rij.size(2)), r=2)
        Rij = Rij[:, :, c]
        R_ij = Rij[:, :, :, 0]
        R_ik = Rij[:, :, :, 1]
        rij = torch.norm(R_ij, dim=3)
        rik = torch.norm(R_ik, dim=3)
        f_rij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        f_rik = smooth_fun(smooth_type=smooth_type, rij=rik, r_inner=R_min, r_outer=R_max)
        cosine = torch.einsum('ijkl,ijkl->ijk', R_ij, R_ik) / (rij * rik)
        # cosine = torch.cos(torch.acos(cosine).unsqueeze(dim=-1) - theta.unsqueeze(dim=1))

        Gi = torch.sum(2 ** (1 - zeta.unsqueeze(dim=1)) * (1 + cosine) ** zeta.unsqueeze(dim=1) * torch.exp(
            -eta.unsqueeze(dim=1) * (0.5 * (rij + rik).unsqueeze(dim=-1) - rs.unsqueeze(dim=1)) ** 2) * (
                                   f_rij * f_rik).unsqueeze(dim=-1), dim=2)
        return Gi

    @staticmethod
    def sym_ang_w(R_min: float,
                  R_max: float,
                  lambd: torch.Tensor,
                  eta: torch.Tensor,
                  zeta: torch.Tensor,
                  rs: torch.Tensor,
                  smooth_type: str,
                  Rij: torch.Tensor):

        Rij = Rij[:, :, :, 1:4]
        c = torch.combinations(torch.arange(Rij.size(2)), r=2)
        Rij = Rij[:, :, c]
        R_ij = Rij[:, :, :, 0]
        R_ik = Rij[:, :, :, 1]
        R_jk = R_ij - R_ik
        rij = torch.norm(R_ij, dim=3)
        rik = torch.norm(R_ik, dim=3)
        rjk = torch.norm(R_jk, dim=3)
        f_rij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        f_rik = smooth_fun(smooth_type=smooth_type, rij=rik, r_inner=R_min, r_outer=R_max)
        f_rjk = smooth_fun(smooth_type=smooth_type, rij=rik, r_inner=R_min, r_outer=R_max)
        cosine = torch.einsum('ijkl,ijkl->ijk', R_ij, R_ik) / (rij * rik)

        Gi = torch.sum(
            2 ** (1 - zeta.unsqueeze(dim=1)) * (1 + lambd.unsqueeze(dim=1) * cosine.unsqueeze(dim=-1)) ** zeta.unsqueeze(
                dim=1) * torch.exp(-eta.unsqueeze(dim=1) * (rij** 2 + rik ** 2 + rjk ** 2).unsqueeze(dim=-1)) * (
                        f_rij * f_rik * f_rjk).unsqueeze(dim=-1), dim=2)
        return Gi



import time
import torch
from collections import OrderedDict
from ase import Atoms
from abfml.param.param import Param
from abfml.logger.loggers import AverageMeter, Logger
from abfml.loss.losser import calculate_weight
from abfml.optimizer.learn_rate import adjust_lr
from abfml.data.read_data import ReadData
import sys


def train_loop(data_load, model, optimizer, iters_step: int, config: Param):
    logger = Logger(config.GlobalSet.logger_file).logger
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device
    batch_size = config.TrainSet.batch_size
    criterion = torch.nn.MSELoss(reduction='sum')
    predict_name = ["energy",  "atomic_energy", "force", "virial"]
    # register losses start
    loss_dict = OrderedDict()
    loss_dict["loss"] = AverageMeter(name="Loss", fmt=".2e", summary_type="AVERAGE")
    loss_dict["time"] = AverageMeter(name="Time", fmt="2.3f", summary_type="SUM")

    loss_set = config.LossSet
    if loss_set.start_energy_weight != 0.0 or loss_set.limit_energy_weight != 0.0:
        loss_dict["energy"] = AverageMeter(name="E_tot", fmt=".2e", summary_type="AVERAGE")
    if loss_set.start_ei_weight != 0.0 or loss_set.limit_ei_weight != 0.0:
        loss_dict["atomic_energy"] = AverageMeter(name="Ei", fmt=".2e", summary_type="AVERAGE")
    if loss_set.start_force_weight != 0.0 or loss_set.limit_force_weight != 0.0:
        loss_dict["force"] = AverageMeter(name="Force", fmt=".2e", summary_type="AVERAGE")
    if loss_set.start_virial_weight != 0.0 or loss_set.limit_virial_weight != 0.0:
        loss_dict["virial"] = AverageMeter(name="Virial", fmt=".2e", summary_type="AVERAGE")
    # register losses end
    weight_dict = {}
    model.train()
    for iters, image_batch in enumerate(data_load):
        time_start = time.time()

        lr_real = adjust_lr(Lr_param=config.LrSet, iters_step=iters_step + iters + 1)
        for param_group in optimizer.param_groups:
            param_group["lr"] = lr_real

        n_atoms = image_batch["n_atoms"][0].int().to(device)
        Rij = image_batch["Rij"].to(dtype=dtype, device=device)
        element_type = image_batch["element_type"][0].int().to(device)
        Zi = image_batch["Zi"].int().to(device)
        Nij = image_batch["Nij"].int().to(device)
        Zij = image_batch["Zij"].int().to(device)

        predict_tulp = model(element_type, Zi, Nij, Zij, Rij, 0)
        Lr_ratio = lr_real / config.LrSet.start_lr
        weight_tulp = calculate_weight(param=config.LossSet, learn_rate=Lr_ratio)
        weight_dict["energy"] = batch_size * n_atoms ** 2
        weight_dict["atomic_energy"] = batch_size * n_atoms
        weight_dict["force"] = batch_size * 3 * n_atoms
        weight_dict["virial"] = batch_size * 9 * n_atoms ** 2

        loss = torch.tensor([0.0], dtype=dtype, device=device)
        loss_val = torch.tensor([0.0], dtype=dtype, device=device)
        for name in loss_dict.keys():
            if name in predict_name:
                try:
                    label_val = image_batch[name].to(dtype=dtype, device=device)
                    indices = predict_name.index(name)
                    predict_val = predict_tulp[indices]
                    MSE = criterion(label_val, predict_val) / weight_dict[name]
                    loss_label_RMSE = torch.sqrt(MSE)
                    loss_val = loss_val + loss_label_RMSE
                    loss = loss + MSE * weight_tulp[indices]
                    loss_dict[name].update(loss_label_RMSE.item(), batch_size)
                except KeyError:
                    raise Exception(f"You are trying to train {name}, "
                                    f"but the dataset doesn't include data for {name}")
        loss.backward()
        optimizer.step()
        optimizer.zero_grad()

        time_end = time.time()
        loss_dict["loss"].update(loss_val.item(), batch_size)
        loss_dict["time"].update(time_end - time_start)
        if iters % config.TrainSet.print_freq == 0:
            logger.info(f"| Train Iters:{iters + 1:>6d} ," + ' ,'.join([str(dit) for dit in loss_dict.values()]))
    logger.info(f"| Train Information: " + ' ,'.join([dit.summary() for dit in loss_dict.values()]))
    return loss_dict


def valid_loop(data_load, model, logger_name: str = 'valid.log', print_freq: int = 1, save_predict: bool = False):
    logger = Logger(logger_name).logger
    dtype = next(model.parameters()).dtype
    device = next(model.parameters()).device
    batch_size = data_load.batch_size
    criterion = torch.nn.MSELoss(reduction='sum')
    predict_name = ["energy", "atomic_energy", "force", "virial"]
    predict_data = {"energy": [], "atomic_energy": [], "force": [], "virial": []}
    # register losses start
    loss_dict = OrderedDict()
    loss_dict["loss"] = AverageMeter(name="Loss", fmt=".2e", summary_type="AVERAGE")
    loss_dict["time"] = AverageMeter(name="Time", fmt="2.3f", summary_type="SUM")
    loss_dict["energy"] = AverageMeter(name="E_tot", fmt=".2e", summary_type="AVERAGE")
    loss_dict["atomic_energy"] = AverageMeter(name="Ei", fmt=".2e", summary_type="AVERAGE")
    loss_dict["force"] = AverageMeter(name="Force", fmt=".2e", summary_type="AVERAGE")
    loss_dict["virial"] = AverageMeter(name="Virial", fmt=".2e", summary_type="AVERAGE")
    # register losses end
    weight_dict = {}
    model.eval()
    common_label = set([])
    for iters, image_batch in enumerate(data_load):
        time_start = time.time()
        n_atoms = image_batch["n_atoms"][0].int().to(device)
        Rij = image_batch["Rij"].to(dtype=dtype, device=device)
        element_type = image_batch["element_type"][0].int().to(device)
        Zi = image_batch["Zi"].int().to(device)
        Nij = image_batch["Nij"].int().to(device)
        Zij = image_batch["Zij"].int().to(device)

        predict_tulp = model(element_type, Zi, Nij, Zij, Rij, 0)

        weight_dict["energy"] = batch_size * n_atoms ** 2
        weight_dict["atomic_energy"] = batch_size * n_atoms
        weight_dict["force"] = batch_size * 3 * n_atoms
        weight_dict["virial"] = batch_size * 9 * n_atoms ** 2

        loss_val = torch.tensor([0], dtype=dtype, device=device)

        common_label = set(loss_dict.keys()).intersection(set(image_batch.keys()))
        if len(common_label) > 0:
            for name in common_label:
                label_val = image_batch[name].to(dtype=dtype, device=device)
                indices = predict_name.index(name)
                predict_val = predict_tulp[indices]
                MSE = criterion(label_val, predict_val) / weight_dict[name]
                loss_label_RMSE = torch.sqrt(MSE)
                loss_val = loss_val + loss_label_RMSE
                loss_dict[name].update(loss_label_RMSE.item(), batch_size)
                if save_predict:
                    predict_data[name].append(predict_val.detach())
        else:
            raise Exception("If you want to verify the accuracy of a dataset, then at least one of the labels "
                            "(energy, atomic_energy, force, stress) should be present in your dataset")
        time_end = time.time()
        loss_dict["loss"].update(loss_val.item(), batch_size)
        loss_dict["time"].update(time_end - time_start)
        if iters % print_freq == 0:
            logger.info(f"| Valid Iters:{iters + 1:>6d} ,"
                        + ' ,'.join([str(loss_dict[label]) for label in ["loss", "time"] + list(common_label)]))
    logger.info(f"| Valid Information: " +
                ' ,'.join([loss_dict[label].summary() for label in ["loss", "time"] + list(common_label)]))
    return loss_dict, predict_data


def predict(atom: Atoms, model):
    type_map = model.type_map
    cutoff = model.cutoff
    neighbor = model.neighbor
    information_tulp = ReadData.calculate_neighbor(atom=atom, cutoff=cutoff, neighbor=neighbor, type_map=type_map)
    element_type, Zi, Nij, Zij, Rij = information_tulp
    predict_tulp = model(element_type, Zi, Nij, Zij, Rij, 0)
    Etot, Ei, Force, Virial, virial = predict_tulp
    return Etot, Ei, Force, Virial, virial



import numpy as np
import matplotlib.pyplot as plt


plt.rcParams.update({'font.size': 18, 'font.family': 'serif', 'font.serif': ['Times New Roman']})
plt.rcParams['mathtext.default'] = 'regular'

plt.figure(figsize=(10, 8))

# 创建数据
x = np.linspace(-5, 5, 100)
y = np.linspace(-5, 5, 100)
X, Y = np.meshgrid(x, y)
Z = np.sin(np.sqrt(X**2 + Y**2))

# 绘制等高线图
contour = plt.contourf(X, Y, Z, levels=20, cmap='viridis')
cbar = plt.colorbar(contour)
cbar.set_label('$ J/m^2 $')
plt.title('Contour Plot')
plt.xlabel('X-axis')
plt.ylabel('Y-axis')
plt.show()



import json
import sys
import importlib.util
import numpy as np
from ase import Atoms
from ase.data import chemical_symbols
from ase.build import bulk
from typing import Optional
from torch.utils.data import ConcatDataset
from abfml.param.param import Param
from abfml.logger.loggers import Logger, log_logo
from abfml.data.read_data import ReadData
from abfml.calculate import ABFML


def check_mlff(
        *,
        model_name: Optional[str] = None,
        input: Optional[str] = None,
        dtype: str = "float32",
        **kwargs):
    logger = Logger("check.log").logger
    log_logo(logger=logger)

    if input is not None:
        with open(input, 'r') as file:
            input_json = json.load(file)
        param_class = Param(input_dict=input_json)
        field_model_name = param_class.user_defined.field_name
        model_path = param_class.user_defined.model_path
        spec = importlib.util.spec_from_file_location(field_model_name, model_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[field_model_name] = module
        spec.loader.exec_module(module)

        field_model = getattr(module, field_model_name)
        if param_class.user_defined.normal_name is not None:
            logger.info(f"| Information: Since you defined Normal, we'll read the train_file as the dataset for Normal")
            train_dataclass = ReadData(filename=param_class.DataSet.train_file,
                                       cutoff=param_class.GlobalSet.cutoff,
                                       neighbor=param_class.GlobalSet.neighbor,
                                       type_map=param_class.GlobalSet.type_map,
                                       file_format=param_class.DataSet.file_format)
            train_data = ConcatDataset(train_dataclass.get_mlffdata())
            normal_class = getattr(module, param_class.user_defined.normal_name)
            normal_model = normal_class(normal_data=train_data, param_class=param_class, normal_rate="auto")
            normal_data = normal_model.normal_data
        else:
            normal_data = None
        model = field_model(type_map=param_class.GlobalSet.type_map,
                            cutoff=param_class.GlobalSet.cutoff,
                            neighbor=param_class.GlobalSet.neighbor,
                            config=param_class.user_defined.config,
                            normal=normal_data)
        calculator = ABFML(model_class=model,dtype=dtype)

    elif model_name is not None:
        calculator = ABFML(model=model_name,dtype=dtype)

    else:
        raise RuntimeError("model_name and input are None?")

    atomic_number = calculator.model.type_map[0]
    element_symbol = chemical_symbols[atomic_number]

    logger.info(f"+----------------------------------- rotate symmetry check ------------------------------------+")
    position = np.array([[15.0, 15.0, 15.0],[13.0, 15.0, 15.0],[15.0, 15.0, 13.0]])
    cluster = Atoms(element_symbol+'3', position)
    cluster.calc = calculator
    energy = [cluster.get_potential_energy()]
    force = [cluster.get_forces()]
    cluster.rotate(45, 'z', center='COP')
    energy.append(cluster.get_potential_energy())
    force.append(cluster.get_forces())
    dF = (np.linalg.norm(force[1], axis=1) - np.linalg.norm(force[0], axis=1)).sum().item()
    logger.info(f"       cluster-> energy: {energy[0]:>.6f}, force-sum: {force[0].sum().item():>.6f}")
    logger.info(f"rotate cluster-> energy: {energy[1]:>.6f}, force-sum: {force[1].sum().item():>.6f}")
    logger.info(f"    difference-> energy: {energy[1]-energy[0]:>.6f}, force: {dF:>.6f}")
    logger.info(f"Information: The magnitude of the energy/and the combined force remains the same after the")
    logger.info(f"             structure is rotated. Isolated system combined force should be 0.")
    logger.warning(f"Part “difference” should be 0. But if it is less than 1e-5, ")
    logger.warning(f"it may be a precision issue, otherwise check the model rotation conservation")
    logger.info(f"+----------------------------------------------------------------------------------------------+")

    logger.info(f"+-------------------------------------- periodicity check -------------------------------------+")
    cluster = bulk('Cu', 'fcc', a=3.62, cubic=True)
    cluster.calc = calculator
    energy = [cluster.get_potential_energy()]
    force = [cluster.get_forces()]
    cluster = cluster * (3, 3, 3)
    cluster.calc = calculator
    energy.append(cluster.get_potential_energy())
    force.append(cluster.get_forces())
    dE = energy[1] - energy[0] * 27
    dF = np.abs(force[1][:4] - force[0]).sum().item()
    logger.info(f"       bulk-> energy: {energy[0]:>.6f}, force: {force[0].sum().item():>.6f}")
    logger.info(f" 3*3*3 bulk-> energy: {energy[1]:>.6f}, force: {force[1].sum().item():>.6f}")
    logger.info(f" difference-> E[1]-E[0]*27: {dE:>.6f},  dF: {dF:>.6f}")
    logger.info(f"Information: The energy of the expanded cell of a periodic structure is proportional to the number")
    logger.info(f"             of times it is expanded and the force on the corresponding atoms is unchanged.")
    logger.warning(f"If the difference is less than 1e-5 it may be an accuracy problem,")
    logger.warning(f"otherwise the energy and force relationship should be checked.")
    logger.info(f"+----------------------------------------------------------------------------------------------+")

    logger.info(f"+------------------------------------- energy-force check -------------------------------------+")
    cluster = Atoms(element_symbol + '3', position)
    cluster.calc = calculator
    energy = [cluster.get_potential_energy()]
    force = [cluster.get_forces()[0]]
    cluster.positions[0] = [15.0, 15.0, 15.002]
    energy.append(cluster.get_potential_energy())
    force.append(cluster.get_forces()[0])
    Fdx = -1 * (force[0][2] + force[1][2]) / 2 * 0.002
    dE = energy[1] - energy[0]
    logger.info(f"       cluster-> energy: {energy[0]:>.6f}, force: {force[0][2].item():>.6f}")
    logger.info(f"  perturbation-> energy: {energy[1]:>.6f}, force: {force[1][2].item():>.6f}")
    logger.info(f"              -> dE: {dE:>.6f}, F * dx: {Fdx:>.6f}, difference: {dE - Fdx:>.6f}")
    logger.info(f"Information: Force and energy should satisfy F=-dE/dx, dE=F*dx for perturbation.")
    logger.warning(f"if the difference is less than 1e-5 it may be an accuracy problem,")
    logger.warning(f"otherwise the energy and force relationship should be checked.")
    logger.info(f"+----------------------------------------------------------------------------------------------+")




import json
import os
import sys
import importlib.util
import time
import torch
import numpy as np
from typing import Optional
from torch.utils.data import ConcatDataset, DataLoader, Subset
from abfml.param.param import Param
from abfml.model.normal import DPNormal, FeatureNormal
from abfml.logger.loggers import Logger, log_data_info, log_logo
from abfml.data.read_data import ReadData
from abfml.train.trainer import train_loop, valid_loop
from abfml.optimizer.optim_init import optim_init
from abfml.optimizer.learn_rate import calc_lr_param
import torch.multiprocessing


def train_mlff(
        *,
        INPUT: str,
        init_model: Optional[str],
        restart: Optional[str],
        **kwargs):
    with open(INPUT, 'r') as file:
        input_json = json.load(file)
    param_class = Param(input_dict=input_json)

    logger = Logger(param_class.GlobalSet.logger_file).logger
    log_logo(logger=logger)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    num_threads = torch.get_num_threads()
    num_worker = int(num_threads / 2)
    # torch.set_num_threads(1)
    logger.info("num_threads : %d" % num_threads)
    torch.set_printoptions(precision=10)
    if param_class.DataSet.ratio is None:
        logger.info("+-------------------------------------- train data file ---------------------------------------+")
        train_dataclass = ReadData(filename=param_class.DataSet.train_file,
                                   cutoff=param_class.GlobalSet.cutoff,
                                   neighbor=param_class.GlobalSet.neighbor,
                                   type_map=param_class.GlobalSet.type_map,
                                   file_format=param_class.DataSet.file_format)
        log_data_info(logger, train_dataclass)
        train_data = ConcatDataset(train_dataclass.get_mlffdata())

        logger.info("+-------------------------------------- valid data file ---------------------------------------+")
        valid_dataclass = ReadData(filename=param_class.DataSet.valid_file,
                                   cutoff=param_class.GlobalSet.cutoff,
                                   neighbor=param_class.GlobalSet.neighbor,
                                   type_map=param_class.GlobalSet.type_map,
                                   file_format=param_class.DataSet.file_format)
        log_data_info(logger, valid_dataclass)
        valid_data = ConcatDataset(valid_dataclass.get_mlffdata())
    else:
        logger.info("+----------------------------------------- data file -------------------------------------------+")
        total_dataclass = ReadData(filename=param_class.DataSet.train_file,
                                   cutoff=param_class.GlobalSet.cutoff,
                                   neighbor=param_class.GlobalSet.neighbor,
                                   type_map=param_class.GlobalSet.type_map,
                                   file_format=param_class.DataSet.file_format)
        log_data_info(logger, total_dataclass)
        total_data = ConcatDataset(total_dataclass.get_mlffdata())
        total_indices = np.arange(len(total_data))
        if param_class.DataSet.shuffle:
            logger.info(f"| Because of shuffle is set to true, ")
            logger.info(f"| The training and validation sets will be randomly selected from all the data.")
            np.random.shuffle(total_indices)
        else:
            logger.info(f"| Because of shuffle is set to false, preserve the order of the dataset, scaled")
        split_point = int(len(total_data) * param_class.DataSet.ratio)
        train_indices, valid_indices = total_indices[:split_point], total_indices[split_point:]
        train_data = Subset(total_data, train_indices)
        valid_data = Subset(total_data, valid_indices)
    train_data_load = DataLoader(train_data, batch_size=param_class.TrainSet.batch_size,
                                 shuffle=True, num_workers=num_worker)
    valid_data_load = DataLoader(valid_data, batch_size=param_class.TrainSet.batch_size,
                                 shuffle=False, num_workers=num_worker)

    if init_model is not None:
        # old_model = torch.jit.script(init_model)
        # if old_model.Rcut != param_class.GlobalSet.cutoff:
        #     raise Exception("")
        # if old_model.neighbor != param_class.GlobalSet.neighbor:
        #     raise Exception("")
        # if param_class.GlobalSet.type_map in old_model.type():
        #     raise Exception("")
        # if param_class.model_name == "deepmd":
        #     if param_class.deep.name == "se_e2_a":
        #         from src.model.dp_se_e2_a import DpSe2a
        #         model = DpSe2a
        # elif param_class.model_name == "user_defined":
        #     model_path = os.path.dirname(param_class.UseDefined.model_path)
        #     sys.path.append(model_path)
        #     try:
        #         import user_model
        #         model = user_model.user_model(param_class.UseDefined.config)
        #     except:
        #         raise Exception(f"Undefined deepmd network: {param_class.deep.name}")
        # else:
        #     raise Exception(f"If you define your own machine learning force field, use the name: user_defined")
        pass

    elif restart is not None:
        check_point_class = torch.load(restart)
        model = check_point_class["model"]
    else:
        if param_class.model_name in ["dp_se_e2_a", "dp_se_e3", "dp_se_e2_r"]:
            if param_class.model_name == "dp_se_e2_r":
                from abfml.model.field_model_dp import DpSe2r
                energy_shift, std_mean = DPNormal(normal_data=train_data,
                                                  param_class=param_class,
                                                  normal_rate='auto',
                                                  coordinate_matrix='r').normal_data
                model_class = DpSe2r
            elif param_class.model_name == "dp_se_e2_a":
                from abfml.model.field_model_dp import DpSe2a
                energy_shift, std_mean = DPNormal(normal_data=train_data,
                                                  param_class=param_class,
                                                  normal_rate='auto',
                                                  coordinate_matrix='a').normal_data
                model_class = DpSe2a
            else:
                from abfml.model.field_model_dp import DpSe3
                energy_shift, std_mean = DPNormal(normal_data=train_data,
                                                  param_class=param_class,
                                                  normal_rate='auto',
                                                  coordinate_matrix='a').normal_data
                model_class = DpSe3

            model = model_class(type_map=param_class.GlobalSet.type_map,
                                cutoff=param_class.GlobalSet.cutoff,
                                neighbor=param_class.GlobalSet.neighbor,
                                fitting_config=param_class.DeepSe.fitting_config,
                                embedding_config=param_class.DeepSe.embedding_config,
                                energy_shift=energy_shift,
                                std_mean=std_mean)

        elif param_class.model_name == "BPMlp":
            from abfml.model.field_model_bp import BPMlp
            energy_shift, std_mean = FeatureNormal(normal_data=train_data,
                                                   param_class=param_class,
                                                   normal_rate='auto',
                                                   feature_name='BP').normal_data
            model = BPMlp(type_map=param_class.GlobalSet.type_map,
                          cutoff=param_class.GlobalSet.cutoff,
                          neighbor=param_class.GlobalSet.neighbor,
                          fit_config=param_class.BPDescriptor.fitting_config,
                          bp_features_information=param_class.BPDescriptor.bp_features_information,
                          bp_features_param=param_class.BPDescriptor.bp_features_param,
                          energy_shift=energy_shift,
                          std_mean=std_mean)

        elif param_class.model_name == "NEP":
            from abfml.model.field_model_nep import NEP
            model = NEP(type_map=param_class.GlobalSet.type_map,
                        cutoff=param_class.GlobalSet.cutoff,
                        neighbor=param_class.GlobalSet.neighbor,
                        fitting_config=param_class.NEPParam.fitting_config,
                        feature_config=param_class.NEPParam.feature_config,
                        energy_shift=[0,0],
                        std_mean=[0])
        elif param_class.model_name == "user_defined":
            field_model_name = param_class.user_defined.field_name
            model_path = param_class.user_defined.model_path
            spec = importlib.util.spec_from_file_location(field_model_name, model_path)
            module = importlib.util.module_from_spec(spec)
            sys.modules[field_model_name] = module
            spec.loader.exec_module(module)

            field_model = getattr(module, field_model_name)
            if param_class.user_defined.normal_name is not None:
                normal_class = getattr(module, param_class.user_defined.normal_name)
                normal_model = normal_class(normal_data=train_data, param_class=param_class, normal_rate="auto")
                normal_data = normal_model.normal_data
            else:
                normal_data = None
            model = field_model(type_map=param_class.GlobalSet.type_map,
                                cutoff=param_class.GlobalSet.cutoff,
                                neighbor=param_class.GlobalSet.neighbor,
                                config=param_class.user_defined.config,
                                normal=normal_data)
        else:
            raise Exception(f"If you define your own machine learning force field, use the name: user_defined")

    if param_class.GlobalSet.dtype == "float32":
        model.to(torch.float32)
    else:
        model.to(torch.float64)
    model.to(device=device)

    if len(param_class.TrainSet.freeze_param) != 0:
        logger.info(f"+----------------------------------- freeze or train param ------------------------------------+")
        namelist = [name for name, _ in model.named_parameters()]
        if not all(name in namelist for name in param_class.TrainSet.freeze_param[1:]):
            raise Exception("Error: Some parameters were not found in the model.")
        if param_class.TrainSet.freeze_param[0] == "freeze":
            logger.info(f"| The following parameters will be frozen.")
            for name, param in model.named_parameters():
                if name in param_class.TrainSet.freeze_param[1:]:
                    param.requires_grad = False
        elif param_class.TrainSet.freeze_param[0] == "train":
            logger.info(f"| The following parameters will be trained.")
            for name, param in model.named_parameters():
                if name not in param_class.TrainSet.freeze_param:
                    param.requires_grad = False
        else:
            raise Exception(f"Keyword {param_class.TrainSet.freeze_param[0]} error, flags only allowed freeze and train!")
        for name in param_class.TrainSet.freeze_param[1:]:
            logger.info(f"| {name:<60s}")
        logger.info(f"+----------------------------------------------------------------------------------------------+")

    iters_step: int = 0
    start_epoch: int = 1
    if restart is not None:
        check_point_class = torch.load(restart)
        optimizer = check_point_class["optimizer"]
        start_epoch = check_point_class["epoch"]
        iters_step = check_point_class["iters_step"]
    else:
        optimizer = optim_init(model=model, param=param_class.TrainSet)
        calc_lr_param(Lr_param=param_class.LrSet,
                      epoch=param_class.TrainSet.epoch,
                      train_iters=int(len(train_data)/param_class.TrainSet.batch_size),
                      logger_name=param_class.GlobalSet.logger_file)

    logger.info(r"########################################--Start  Train--########################################")
    logger.info(f"+------------------------------------------Start Loop------------------------------------------+")
    time_loop_start = time.time()
    for epoch in range(start_epoch, param_class.TrainSet.epoch + 1):
        logger.info(f"|-- Train Epoch:{epoch:>3d}, batch size:{train_data_load.batch_size:>3d}, "
                    f"total iters: {len(train_data_load):>6d} --------------------------------------|")
        train_loss_dict = train_loop(data_load=train_data_load,
                                     model=model,
                                     optimizer=optimizer,
                                     iters_step=iters_step,
                                     config=param_class)

        iters_step = iters_step + len(train_data_load)

        logger.info(f"|-- Valid Epoch:{epoch:>3d}, batch size:{valid_data_load.batch_size:>3d}, "
                    f"total iters: {len(valid_data_load):>6d}                                       |")
        valid_loss_dict, _ = valid_loop(data_load=valid_data_load,
                                        model=model,
                                        logger_name=param_class.GlobalSet.logger_file,
                                        print_freq=param_class.TrainSet.print_freq,
                                        save_predict=False)

        # ************************************** Start writing loss file ************************************** #
        if not os.path.exists(param_class.LossSet.loss_file) or epoch == 1:
            header = '{:>8}'.format("Epoch")
            header_time = ""
            for train_loss_key in train_loss_dict.keys():
                if train_loss_key == "time":
                    header_time = '{:>16s}(s)'.format("Total_Time")
                else:
                    header += '{:>18s}{:>18s}'.format("T_RMSE_" + train_loss_dict[train_loss_key].name,
                                                      "V_RMSE_" + train_loss_dict[train_loss_key].name)
            header += header_time
            with open(param_class.LossSet.loss_file, 'w') as fp:
                fp.write(header)

        summaries = f"\n{epoch:>8d}"
        summaries_time = ""
        for train_loss_key in train_loss_dict.keys():
            if train_loss_key == "time":
                total_time = '{:6.2f}'.format(train_loss_dict["time"].sum + valid_loss_dict["time"].sum)
                summaries_time = '{:>16s}'.format(total_time)
            else:
                summaries += '{:>18s}'.format(format(train_loss_dict[train_loss_key].avg, '.5e'))
                summaries += '{:>18s}'.format(format(valid_loss_dict[train_loss_key].avg, '.5e'))
        summaries += summaries_time
        with open(param_class.LossSet.loss_file, 'a') as fp:
            fp.write(summaries)
        # **************************************  Stop writing loss file ************************************** #

        if epoch % param_class.TrainSet.save_check == 0:
            check_point_path = "checkpoint_" + str(epoch) + ".ckpt"
        elif epoch == param_class.TrainSet.epoch:
            check_point_path = "checkpoint_final.ckpt"
        else:
            check_point_path = "checkpoint_latest.ckpt"

        torch.save({
            'epoch': epoch,
            'iters_step': iters_step,
            'model': model,
            'optimizer': optimizer
        }, check_point_path)
    try:
        jit_model = torch.jit.script(model)
        jit_model.save(param_class.TrainSet.model_file)
    except Exception as e:
        torch.save(model, param_class.TrainSet.model_file+'.pth')
        logger.info(f"| Information: Failed to save the model via torchscript, will save the model by torch.save! \n"
                    f"               The model can't be loaded in lammps, but it can be used in ASE")
    time_loop_end = time.time()
    logger.info(f"+------------------------------------------ End Loop ------------------------------------------+")
    logger.info(f"| Total training time:{time_loop_end - time_loop_start:8.2f}s, total training iters:{iters_step:>8d}")
    logger.info(r"######################################## --End  Train-- ########################################")

    return



import time
import torch
import torch.nn as nn
from abfml.model.method import FieldModel
from typing import List, Optional, Tuple
from abfml.model.math_fun import smooth_fun
from abfml.model.network import EmbeddingNet, FittingNet
from typing import List, Optional, Dict, Any


@torch.jit.interface
class ModuleInterface(torch.nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass


class DpSe2a(FieldModel):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int],
                 fitting_config: Dict[str, Any],
                 embedding_config: Dict[str, Any],
                 energy_shift: List[float],
                 std_mean: Optional[List[torch.Tensor]]):
        super(DpSe2a, self).__init__(type_map=type_map, cutoff=cutoff, neighbor=neighbor)
        self.M2 = embedding_config["M2"]
        self.R_min = embedding_config["R_min"]
        self.R_max = embedding_config["R_max"]
        self.smooth_fun = embedding_config['smooth_fun']

        self.embedding_net = nn.ModuleList()
        self.fitting_net = nn.ModuleList()
        self.embedding_net_index = []
        self.fitting_net_index = []
        self.embedding_size = embedding_config["network_size"]
        self.std_mean = std_mean

        for i, itype in enumerate(self.type_map):
            for j, jtype in enumerate(self.type_map):
                self.embedding_net.append(EmbeddingNet(network_size=embedding_config["network_size"],
                                                       activate=embedding_config["activate_function"],
                                                       bias=embedding_config["bias"],
                                                       resnet_dt=embedding_config["resnet_dt"]))
                self.embedding_net_index.append(str(itype) + "_" + str(jtype))
            fit_network_size = [embedding_config["network_size"][-1] * embedding_config["M2"]] + fitting_config[
                "network_size"]
            self.fitting_net.append(FittingNet(network_size=fit_network_size,
                                               activate=fitting_config["activate_function"],
                                               bias=fitting_config["bias"],
                                               resnet_dt=fitting_config["resnet_dt"],
                                               energy_shift=energy_shift[i]))
            self.fitting_net_index.append(str(itype))

    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              Zij: torch.Tensor,
              Nij: torch.Tensor,
              Rij: torch.Tensor,
              n_ghost: int) -> Tuple[torch.Tensor, torch.Tensor]:
        # t1 = time.time()
        batch, n_atoms, max_neighbor = Zij.shape
        device = Rij.device
        dtype = Rij.dtype
        type_map: List[int] = element_map.to(torch.int64).tolist()
        width_map: List[int] = [0]
        for ii in type_map:
            width_map.append(self.neighbor[self.type_map.index(ii)] + width_map[-1])
        # t2 = time.time()
        # Ri[batch, n_atoms, max_neighbor, 4] 4-->(srij, srij * xij/rij, srij * zij/rij, srij * zij/rij)
        Ri = DpSe2a.calculate_coordinate_matrix(R_min=self.R_min, R_max=self.R_max,
                                                smooth_type=self.smooth_fun, Rij=Rij)
        Ri = DpSe2a.scale(self.type_map, type_map, self.std_mean, Ri, Zi)

        # Ei[batch, n_atoms, 1]
        Ei = torch.zeros(batch, n_atoms, 1, dtype=dtype, device=device)
        for i, itype in enumerate(type_map):
            mask_itype = (Zi == itype)
            if not mask_itype.any():
                continue
            i_Ri = Ri[mask_itype].reshape(batch, -1, max_neighbor, 4)
            RiT_G = torch.zeros(batch, i_Ri.shape[1], 1, self.embedding_size[-1], dtype=dtype, device=device)

            for j, jtype in enumerate(type_map):
                # srij[batch, n_atoms of itype, max_neighbor, 1]
                srij = i_Ri[:, :, width_map[j]:width_map[j+1], 0].unsqueeze(-1)
                ij = self.embedding_net_index.index(str(itype) + "_" + str(jtype))
                embedding_net: ModuleInterface = self.embedding_net[ij]
                # G[batch, n_atoms of itype, max_neighbor, embedding_net_size[-1]]
                G = embedding_net.forward(srij)
                RiT = i_Ri[:, :, width_map[j]:width_map[j+1], :].transpose(-2, -1)
                temp_b = torch.matmul(RiT, G)
                RiT_G = RiT_G + temp_b

            RiT_G = RiT_G / max_neighbor
            RiT_GM = RiT_G[:, :, :, :self.M2]
            # Di[batch, n_atoms of itype, M2 * embedding_net_size[-1]]
            Di = torch.matmul(RiT_G.transpose(-2, -1), RiT_GM).reshape(batch, i_Ri.shape[1], -1)
            ii = self.fitting_net_index.index(str(itype))
            fitting_net: ModuleInterface = self.fitting_net[ii]
            # Ei_mask[batch, n_atoms of itype, 1]
            Ei_mask = fitting_net.forward(Di)
            Ei[mask_itype] = Ei_mask.reshape(-1, 1)

        Etot = torch.sum(Ei, dim=1)
        return Etot, Ei

    @staticmethod
    def calculate_coordinate_matrix(R_min: float,
                                    R_max: float,
                                    smooth_type: str,
                                    Rij: torch.Tensor) -> torch.Tensor:
        batch, n_atoms, max_neighbor, _ = Rij.shape
        device = Rij.device
        dtype = Rij.dtype
        rij = Rij[:, :, :, 0]
        xij = Rij[:, :, :, 1]
        yij = Rij[:, :, :, 2]
        zij = Rij[:, :, :, 3]
        rij.requires_grad_()

        # Guaranteed not to divide by 0
        mask_rij = (rij > 1e-5)
        rr = torch.zeros(rij.shape, dtype=dtype, device=device)
        rr[mask_rij] = 1 / rij[mask_rij]
        r2 = torch.zeros(rij.shape, dtype=dtype, device=device)
        r2[mask_rij] = 1 / rij[mask_rij] ** 2

        Srij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)

        # Ri[batch, n_atoms, max_neighbor, 4] 4-->[Srij, Srij * xij / rij, Srij * yij / rij, Srij * yij / rij]
        Ri = torch.zeros(batch, n_atoms, max_neighbor, 4, dtype=dtype, device=device)
        Ri[:, :, :, 0] = Srij
        Ri[:, :, :, 1] = Srij * xij * rr
        Ri[:, :, :, 2] = Srij * yij * rr
        Ri[:, :, :, 3] = Srij * zij * rr

        # t3 = time.time()
        # print(t3-t2, t2-t1, t3-t1)
        return Ri

    @staticmethod
    def scale(type_map_all: List[int],
              type_map_use: List[int],
              std_mean: List[torch.Tensor],
              Ri: torch.Tensor,
              Zi: torch.Tensor) -> torch.Tensor:
        for i, element in enumerate(type_map_use):
            indices = type_map_all.index(element)
            device = Ri.device
            mask = (Zi == element)
            std = std_mean[0][indices].detach().to(device)
            avg = std_mean[1][indices].detach().to(device)
            Ri[mask] = (Ri[mask] - avg) / std
        return Ri


class DpSe2r(FieldModel):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int],
                 fitting_config: Dict[str, Any],
                 embedding_config: Dict[str, Any],
                 energy_shift: List[float],
                 std_mean: Optional[List[torch.Tensor]]):
        super(DpSe2r, self).__init__(type_map=type_map, cutoff=cutoff, neighbor=neighbor)
        self.type_map = type_map
        self.neighbor = neighbor
        self.cutoff = cutoff
        self.M2 = embedding_config["M2"]
        self.R_min = embedding_config["R_min"]
        self.R_max = embedding_config["R_max"]
        self.smooth_fun = embedding_config['smooth_fun']

        self.embedding_net = nn.ModuleList()
        self.fitting_net = nn.ModuleList()
        self.embedding_net_index = []
        self.fitting_net_index = []
        self.embedding_size = embedding_config["network_size"]
        self.std_mean = std_mean

        for i, itype in enumerate(self.type_map):
            for j, jtype in enumerate(self.type_map):
                self.embedding_net.append(EmbeddingNet(network_size=embedding_config["network_size"],
                                                       activate=embedding_config["activate_function"],
                                                       bias=embedding_config["bias"],
                                                       resnet_dt=embedding_config["resnet_dt"]))
                self.embedding_net_index.append(str(itype) + "_" + str(jtype))
            fit_network_size = [embedding_config["network_size"][-1] * embedding_config["M2"]] + fitting_config[
                "network_size"]
            self.fitting_net.append(FittingNet(network_size=fit_network_size,
                                               activate=fitting_config["activate_function"],
                                               bias=fitting_config["bias"],
                                               resnet_dt=fitting_config["resnet_dt"],
                                               energy_shift=energy_shift[i]))
            self.fitting_net_index.append(str(itype))

    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              neighbor_list: torch.Tensor,
              Zij: torch.Tensor,
              Rij: torch.Tensor,
              n_ghost: int):
        # time1 = time.time()
        batch, n_atoms, max_neighbor = neighbor_list.shape
        device = Rij.device
        dtype = Rij.dtype
        # Must be set to int64, or it will cause a type mismatch when run in c++.
        type_map = element_map.to(torch.int64).tolist()
        width_map: List[int] = [0]
        for ii in type_map:
            width_map.append(self.neighbor[self.type_map.index(ii)] + width_map[-1])
        # Ri[batch, n_atoms, max_neighbor,1] 1-->(srij) For se_r consider only Srij

        Ri = DpSe2r.calculate_coordinate_matrix(R_min=self.R_min, R_max=self.R_max,
                                                smooth_type=self.smooth_fun, Rij=Rij)
        Ri = DpSe2r.scale(self.type_map, type_map, self.std_mean, Ri, Zi)
        # Ei[batch, n_atoms, 1]
        Ei = torch.zeros(batch, n_atoms, 1, dtype=dtype, device=device)
        for i, itype in enumerate(type_map):
            mask_itype = (Zi == itype)
            if not mask_itype:
                continue
            i_Ri = Ri[mask_itype].reshape(batch, -1, max_neighbor, 1)
            Gi = torch.zeros(batch, i_Ri.shape[1], max_neighbor, self.embedding_size[-1], dtype=dtype, device=device)
            for j, jtype in enumerate(type_map):
                # srij[batch, n_atoms of itype, max_neighbor, 1]
                srij = i_Ri[:, :, width_map[j]:width_map[j+1], 0].unsqueeze(-1)
                ij = self.embedding_net_index.index(str(itype) + "_" + str(jtype))
                embedding_net: ModuleInterface = self.embedding_net[ij]
                # G[batch, n_atoms of itype, max_neighbor, embedding_net_size[-1]]
                Gj = embedding_net.forward(srij)
                Gi[:, :, width_map[j]:width_map[j+1], :] = Gj
            G = torch.einsum('bijm,bikm->bim', Gi, Gi)
            G = G / max_neighbor ** 2
            Di = G.reshape(batch, i_Ri.shape[1], -1)
            # Di[batch, n_atoms of itype, embedding_net_size[-1]]
            ii = self.fitting_net_index.index(str(itype))
            fitting_net: ModuleInterface = self.fitting_net[ii]
            # Ei_mask[batch, n_atoms of itype, 1]
            Ei_mask = fitting_net.forward(Di)
            Ei[mask_itype] = Ei_mask.reshape(-1, 1)
        # Etot[batch, 1]
        Etot = torch.sum(Ei, dim=1)

        return Etot, Ei

    @staticmethod
    def calculate_coordinate_matrix(R_min: float,
                                    R_max: float,
                                    smooth_type: str,
                                    Rij: torch.Tensor) -> torch.Tensor:
        batch, n_atoms, max_neighbor, _ = Rij.shape
        device = Rij.device
        dtype = Rij.dtype

        rij = Rij[:, :, :, 0]
        Srij = smooth_fun(smooth_type=smooth_type, rij=rij, r_inner=R_min, r_outer=R_max)
        # Ri [batch, n_atoms, max_neighbor * len(type_map), 1]
        Ri = torch.zeros(batch, n_atoms, max_neighbor, 1, dtype=dtype, device=device)
        Ri[:, :, :, 0] = Srij

        return Ri

    @staticmethod
    def scale(type_map_all: List[int],
              type_map_use: List[int],
              std_mean: List[torch.Tensor],
              Ri: torch.Tensor,
              Zi: torch.Tensor) -> torch.Tensor:
        for i, element in enumerate(type_map_use):
            indices = type_map_all.index(element)
            device = Ri.device
            mask = (Zi == element)
            std = std_mean[0][indices].detach().to(device)
            avg = std_mean[1][indices].detach().to(device)
            Ri[mask] = (Ri[mask] - avg[0]) / std[0]
        return Ri


class DpSe3(FieldModel):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int],
                 fitting_config: Dict[str, Any],
                 embedding_config: Dict[str, Any],
                 energy_shift: List[float],
                 std_mean: Optional[List[torch.Tensor]]):
        super(DpSe3, self).__init__(type_map=type_map, cutoff=cutoff, neighbor=neighbor)
        self.type_map = type_map
        self.neighbor = neighbor
        self.cutoff = cutoff
        self.M2 = embedding_config["M2"]
        self.R_min = embedding_config["R_min"]
        self.R_max = embedding_config["R_max"]
        self.smooth_fun = embedding_config['smooth_fun']

        self.embedding_net = nn.ModuleList()
        self.fitting_net = nn.ModuleList()
        self.embedding_net_index = []
        self.fitting_net_index = []
        self.embedding_size = embedding_config["network_size"]
        self.std_mean = std_mean

        for i, itype in enumerate(self.type_map):
            for j, jtype in enumerate(self.type_map):
                self.embedding_net.append(EmbeddingNet(network_size=embedding_config["network_size"],
                                                       activate=embedding_config["activate_function"],
                                                       bias=embedding_config["bias"],
                                                       resnet_dt=embedding_config["resnet_dt"]))
                self.embedding_net_index.append(str(itype) + "_" + str(jtype))
            fit_network_size = [self.embedding_size[-1]] + fitting_config[
                "network_size"]
            self.fitting_net.append(FittingNet(network_size=fit_network_size,
                                               activate=fitting_config["activate_function"],
                                               bias=fitting_config["bias"],
                                               resnet_dt=fitting_config["resnet_dt"],
                                               energy_shift=energy_shift[i]))
            self.fitting_net_index.append(str(itype))

    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              neighbor_list: torch.Tensor,
              Zij: torch.Tensor,
              Rij: torch.Tensor,
              n_ghost: int):
        batch, n_atoms, max_neighbor = neighbor_list.shape
        device = Rij.device
        dtype = Rij.dtype
        # Must be set to int64, or it will cause a type mismatch when run in c++.
        type_map: List[int] = element_map.to(torch.int64).tolist()
        width_map: List[int] = [0]
        for ii in type_map:
            width_map.append(self.neighbor[self.type_map.index(ii)] + width_map[-1])

        # Ri[batch, n_atoms, max_neighbor, 4] 4-->(srij, srij * xij/rij, srij * zij/rij, srij * zij/rij)
        Ri = DpSe2a.calculate_coordinate_matrix(R_min=self.R_min, R_max=self.R_max,
                                                smooth_type=self.smooth_fun, Rij=Rij)
        Ri = DpSe2a.scale(self.type_map, type_map, self.std_mean, Ri, Zi)

        rij = Rij[:, :, :, 0].unsqueeze(-1)
        # Guaranteed not to divide by 0
        mask_rij = (rij > 1e-5)
        rr = torch.zeros(rij.shape, dtype=dtype, device=device)
        rr[mask_rij] = 1 / rij[mask_rij]
        # theta_ijk [batch, n_atoms, max_neighbor, max_neighbor]
        theta_ijk = (Ri[:, :, :, 0].unsqueeze(-1) * Ri[:, :, :, 0].unsqueeze(-2)
                     * (torch.matmul(Rij[:, :, :, 1:], Rij[:, :, :, 1:].transpose(-1, -2))
                     * (rr * rr.transpose(-2, -1))))

        # Ei[batch, n_atoms, 1]
        Ei = torch.zeros(batch, n_atoms, 1, dtype=dtype, device=device)
        for i, itype in enumerate(type_map):

            mask_itype = (Zi == itype)
            if not mask_itype.any():
                continue
            i_theta_ijk = theta_ijk[mask_itype].reshape(batch, -1, max_neighbor, max_neighbor, 1)
            Gi = torch.zeros(batch, i_theta_ijk.shape[1], max_neighbor, max_neighbor, self.embedding_size[-1],
                             dtype=dtype, device=device)

            for j, jtype in enumerate(type_map):
                theta = i_theta_ijk[:, :, width_map[j]:width_map[j+1], :, 0].unsqueeze(-1)
                ij = self.embedding_net_index.index(str(itype) + "_" + str(jtype))
                embedding_net: ModuleInterface = self.embedding_net[ij]
                # G[batch, n_atoms of itype, max_neighbor,max_neighbor, embedding_net_size[-1]]
                Gj = embedding_net.forward(theta)
                Gi[:, :, width_map[j]:width_map[j + 1], :, :] = Gj
            # Di[batch, n_atoms of itype, embedding_size[-1]]
            Di = (torch.mul(i_theta_ijk, Gi)).sum(dim=[2, 3]) / max_neighbor ** 2
            ii = self.fitting_net_index.index(str(itype))
            fitting_net: ModuleInterface = self.fitting_net[ii]
            # Ei_mask[batch, n_atoms of itype, 1]
            Ei_mask = fitting_net.forward(Di)
            Ei[mask_itype] = Ei_mask.reshape(-1, 1)
        Etot = torch.sum(Ei, dim=1)
        return Etot, Ei



import ase
import torch
import numpy as np
from typing import List
from ase.calculators.calculator import Calculator, all_changes, PropertyNotImplementedError
from abfml.data.read_data import ReadData


class ABFML(Calculator):
    implemented_properties = ['energy', 'energies', 'forces', 'stress', 'stresses']

    def __init__(self, model: str = None, dtype: str = 'float64', model_class = None, **kwargs) -> None:
        Calculator.__init__(self, **kwargs)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if model is not None:
            try:
                self.model = torch.jit.load(model)
            except RuntimeError:
                try:
                    self.model = torch.load(model)
                except Exception as e:
                    raise RuntimeError(f"Failed to load model using both jit and torch.load: {e}")
        elif model_class is not None:
            self.model = model_class
        else:
            raise RuntimeError('Please give the model file')

        if dtype == "float32":
            self.dtype = torch.float32
        else:
            self.dtype = torch.float64

        self.model.to(self.dtype).to(device=self.device)
        self.model.eval()

    def calculate(self,
                  atoms: ase.Atoms = None,
                  properties: List[str] = None,
                  system_changes: List[str] = all_changes) -> None:

        type_map = self.model.type_map
        cutoff = self.model.cutoff
        neighbor = self.model.neighbor
        information_tulp = ReadData.calculate_neighbor(atom=atoms, cutoff=cutoff, neighbor=neighbor, type_map=type_map)
        element_type, Zi, Nij, Zij, Rij = information_tulp
        predict_tulp = self.model(torch.tensor(element_type),
                                  torch.tensor(Zi).to(device=self.device),
                                  torch.tensor(Nij).to(device=self.device),
                                  torch.tensor(Zij).to(device=self.device),
                                  torch.tensor(Rij).to(self.dtype).to(device=self.device), 0)
        self.results["energy"] = predict_tulp[0].detach().numpy().item()
        self.results["energies"] = predict_tulp[1].detach().numpy().reshape(-1)
        self.results["forces"] = predict_tulp[2].detach().numpy().reshape(-1, 3)
        if "stress" in properties:
            stress = predict_tulp[3].detach().numpy().reshape(3, 3)
            if any(atoms.get_pbc()):
                stress = -0.5 * (stress * stress.T) / atoms.get_volume()
                self.results["stress"] = stress

        if "stresses" in properties:
            stresses = predict_tulp[4].detach().numpy().reshape(-1, 3, 3)
            if any(atoms.get_pbc()):
                self.results["stresses"] = -1.0 * stresses / atoms.get_volume()
            else:
                raise PropertyNotImplementedError




import torch
from abfml.param.param import TrainSet


def optim_init(model, param: TrainSet):
    if param.optimizer == "Adam":
        optimizer = torch.optim.Adam(model.parameters(), eps=1e-5)
    elif param.optimizer == "user_defined":
        optimizer = None
    else:
        raise Exception("")
    return optimizer







import os
import sys
from abfml.entrypoints.main_parser import main_parse_args
from abfml.entrypoints.main_command import main_command
from abfml.entrypoints.train import train_mlff
from abfml.entrypoints.valid import valid_mlff


def main():
    args = main_parse_args()
    main_command(args=args)


if __name__ == '__main__':
    print("__main__")
    train_mlff(INPUT=r"D:\Work\PyCharm\ABFML\example\new-MLFF\usedefined_input.json", init_model=None, restart=None)
    # valid_mlff(model=r"D:\work\Pycharm\mlff\test\checkpoint_final.ckpt", numb_test=100,plot=False,
    #            shuffle=False,
    #            datafile=r"D:\work\Pycharm\mlff\test\data\Cu_128_1200_vasprun.xml")


import sys
import pandas as pd
import matplotlib.pyplot as plt


if __name__ == '__main__':
    loss_file = sys.argv[1]

    data_pd = pd.read_csv(loss_file, sep=r'\s+', header=0)

    label_list = ["Loss", "E_tot", "Force", "Ei", "Virial"]
    for label in label_list:
        if "T_RMSE_" + label in data_pd.keys():
            plt.plot(data_pd["Epoch"], data_pd["T_RMSE_" + label], label='train')
            plt.plot(data_pd["Epoch"], data_pd["V_RMSE_" + label], label='valid')
            plt.xlabel('epoch')
            plt.ylabel(label)
            plt.grid(linestyle='--')
            plt.legend()
            plt.savefig(f'loss_{label}.png', dpi=128)
            plt.close()



import math
from abfml.param.param import LrSet
from abfml.logger.loggers import Logger


def adjust_lr(Lr_param: LrSet, iters_step: int):
    if Lr_param.type_name == 'exp':
        lr_t = Lr_param.start_lr * Lr_param.decay_rate ** (iters_step / Lr_param.decay_step)
    elif Lr_param.type_name == 'exp_decay':
        lr_t = Lr_param.start_lr * Lr_param.decay_rate ** int(iters_step / Lr_param.decay_step)
    else:
        raise TypeError(f'Unsupported type_name of learning rate: {Lr_param.type_name}')
    if lr_t < Lr_param.limit_lr:
        lr_t = Lr_param.limit_lr
    return lr_t


def calc_lr_param(Lr_param: LrSet, epoch: int, train_iters: int, logger_name: str = None):
    total_batches = epoch * train_iters
    if Lr_param.decay_rate is None:
        Lr_param.decay_rate = (Lr_param.limit_lr / Lr_param.start_lr) ** (Lr_param.decay_step / total_batches)
        if logger_name is not None:
            logger = Logger(logger_name).logger
            logger.info(f'start learn rate: {Lr_param.start_lr}, limit learn rate:{Lr_param.limit_lr}, '
                        f'decay step:{Lr_param.decay_step}, decay rate will be: {Lr_param.decay_rate}')
    else:
        need_batches = math.log(Lr_param.limit_lr / Lr_param.start_lr) / math.log(Lr_param.decay_rate) * Lr_param.decay_step
        if logger_name is not None:
            logger = Logger(logger_name).logger
            if need_batches < total_batches:
                logger.info(f'Will reach limit_lr in about {math.ceil(need_batches/train_iters)} epochs.')
            else:
                logger.warning(f'Need {math.ceil(need_batches/train_iters)} epochs to reach limit_lr, '
                               f'but only train for {epoch} epochs!')




import torch
import torch.nn as nn
import numpy as np
from torch.utils.data import DataLoader, Subset
from abfml.model.math_fun import compute_forces_and_virial, calculate_distance_grad
from abfml.param.param import Param
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union


class FieldModel(nn.Module, ABC):
    def __init__(self,
                 type_map: List[int],
                 cutoff: float,
                 neighbor: List[int]):
        super(FieldModel, self).__init__()
        self.type_map = type_map
        self.neighbor = neighbor
        self.cutoff = cutoff

    def forward(self,
                element_map: torch.Tensor,
                Zi: torch.Tensor,
                Nij: torch.Tensor,
                Zij: torch.Tensor,
                Rij: torch.Tensor,
                n_ghost: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        batch, n_atoms, max_neighbor = Nij.shape
        device = Rij.device
        dtype = Rij.dtype
        Ri, dRi = calculate_distance_grad(Ri=Rij)
        Ri.requires_grad_(True)
        Etot, Ei = self.field(element_map=element_map,
                              Zi=Zi,
                              Nij=Nij,
                              Zij=Zij,
                              Rij=Ri,
                              n_ghost=n_ghost)
        if Etot.shape != (batch, 1):
            raise ValueError(" Etot must be torch.Tensor and have shape (batch, 1)")
        if Ei.shape != (batch, n_atoms, 1):
            raise ValueError(" Ei must be torch.Tensor and have shape (batch, n_atoms, 1)")
        mask: List[Optional[torch.Tensor]] = [torch.ones_like(Ei)]
        dE = torch.autograd.grad([Ei], [Ri], grad_outputs=mask, retain_graph=True, create_graph=True)[0]
        assert dE is not None
        dE = torch.unsqueeze(dE, dim=-1)
        dE_Rid = -1.0 * torch.mul(dE, dRi).sum(dim=-2)
        Force, Virial, virial = compute_forces_and_virial(dE_Rid=dE_Rid,
                                                          Rij=Rij, Nij=Nij,
                                                          n_ghost=n_ghost, dtype=dtype, device=device)
        # print(Etot, Force, Rij[0,0,:10])
        return Etot, Ei, Force, Virial, virial

    @abstractmethod
    def field(self,
              element_map: torch.Tensor,
              Zi: torch.Tensor,
              Nij: torch.Tensor,
              Zij: torch.Tensor,
              Rij: torch.Tensor,
              n_ghost: int) -> Tuple[torch.Tensor, torch.Tensor]:
        pass


class NormalModel:
    def __init__(self,
                 normal_data,
                 param_class: Param,
                 normal_rate: Union[float, str] = 'auto',
                 is_get_energy_shift: bool = False):
        self.param_class = param_class
        self.normal_loader = NormalModel.normal_data_loader(need_data=normal_data, normal_rate=normal_rate)
        self.is_get_energy_shift = is_get_energy_shift
        self.normal_data = tuple([])

    def initialize(self):
        normal_data = self.normal(normal_loader=self.normal_loader, param_class=self.param_class)

        if isinstance(normal_data, tuple):
            self.normal_data = normal_data
        else:
            self.normal_data = tuple([normal_data])

        if self.is_get_energy_shift:
            energy_shift = NormalModel.get_energy_shift(need_data=self.normal_loader, type_map=self.param_class.GlobalSet.type_map)
            self.normal_data = tuple([energy_shift]) + self.normal_data

    @staticmethod
    def normal_data_loader(need_data, normal_rate: Union[float, str]) -> DataLoader:
        total_image_num = len(need_data)
        total_indices = np.arange(total_image_num)
        if isinstance(normal_rate, float):
            if normal_rate <= 1.0:
                num = int(total_image_num * normal_rate + 1)
            else:
                raise Exception("rate")
        elif normal_rate == "auto":
            if total_image_num * 0.1 < 100:
                num = total_image_num
            else:
                num = int(total_image_num * 0.1 + 1)
        else:
            raise Exception("rate")
        np.random.shuffle(total_indices)
        normal_indices = total_indices[:num]
        normal_data = Subset(need_data, normal_indices)
        num_threads = torch.get_num_threads()
        num_worker = int(num_threads / 2)
        normal_data_loader = DataLoader(normal_data, batch_size=1, shuffle=True, num_workers=num_worker)

        return normal_data_loader

    @staticmethod
    def get_energy_shift(need_data, type_map: List[int]) -> List[float]:
        ntype = len(type_map)
        type_num = torch.zeros(ntype)
        energy_shift = [0.0] * ntype
        for i, image_batch in enumerate(need_data):
            Zi = image_batch["Zi"]
            element_type = image_batch["element_type"][0].to(torch.int64).tolist()
            for i_type, element in enumerate(element_type):
                mask = (Zi == element)
                indices = type_map.index(element)
                type_num[indices] += 1
                try:
                    energy = torch.mean(image_batch["energy"] / image_batch["n_atoms"]).item()
                    energy_shift[indices] = energy_shift[indices] + energy
                except KeyError:
                    try:
                        Ei = torch.mean(image_batch["atomic_energy"][mask]).item()
                        energy_shift[indices] = energy_shift[indices] + Ei
                    except KeyError:
                        energy_shift[indices] = energy_shift[indices] + np.random.uniform(-10.0, 0.0)

        type_num[type_num == 0] = 1
        for i, i_energy in enumerate(energy_shift):
            energy_shift[i] = (i_energy / type_num[i]).item()

        return energy_shift

    def normal(self, normal_loader, param_class):
        return None





import json
import torch
import numpy as np
from torch.utils.data import ConcatDataset, DataLoader, Subset
from typing import Optional
from abfml.logger.loggers import Logger, log_data_info, log_logo
from abfml.data.read_data import ReadData
from abfml.train.trainer import valid_loop


def valid_mlff(
        *,
        model: str,
        numb_test: int,
        plot: bool,
        shuffle: bool,
        datafile: Optional[list[str]],
        **kwargs):
    logger = Logger("valid.log").logger
    log_logo(logger=logger)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    num_threads = torch.get_num_threads()
    num_worker = int(num_threads / 4)

    if ".ckpt" in model:
        logger.info(f"| Because file {model} has the suffix ckpt,")
        logger.info(f"| we will attempt to read the model from a checkpoint-type file.")
        model_class = torch.load(model)["model"]
    else:
        logger.info(f"| We will attempt to read the model from a jit_script-type file {model}.")
        model_class = torch.jit.load(model)
    model_class.to(device=device)

    if 'valid.input' in datafile[0] and len(datafile) == 1:
        with open('valid.input', 'r') as file:
            input_json = json.load(file)
            filename_list = input_json['valid_file']
    else:
        filename_list = datafile

    logger.info("+-------------------------------------- valid data file ---------------------------------------+")
    valid_dataclass = ReadData(filename=filename_list,
                               cutoff=model_class.cutoff,
                               neighbor=model_class.neighbor,
                               type_map=model_class.type_map,
                               file_format=None)
    log_data_info(logger, valid_dataclass)
    valid_data = ConcatDataset(valid_dataclass.get_mlffdata())
    total_indices = np.arange(len(valid_data))
    if shuffle:
        logger.info(f"| You will randomly select  {numb_test:>4d} image")
        np.random.shuffle(total_indices)
    else:
        logger.info(f"| You will use the first {numb_test:>4d} image")
    logger.info(f"+----------------------------------------------------------------------------------------------+")
    valid_indices = total_indices[:numb_test]
    subset = Subset(valid_data, valid_indices)
    valid_data_load = DataLoader(subset, batch_size=1, shuffle=False, num_workers=0)
    _, predict_data = valid_loop(data_load=valid_data_load,
                                 model=model_class,
                                 logger_name="valid.log",
                                 print_freq=1,
                                 save_predict=True)
    logger.info(f"+----------------------------------------------------------------------------------------------+")
    unit = {'energy': 'eV', 'force':'eV/$\AA$', 'virial': 'eV'}
    # Print the Quality Of Data header
    logger.info(f"|{'Quality Of Data'.center(94)}|")
    logger.info(f"| {'Key':<16} {'DFT Mean':>17} {'DFT Sigma':>17} {'Predict Mean':>17} {'Predict Sigma':>17}     |")
    logger.info(f"+----------------------------------------------------------------------------------------------+")

    # Store evaluation metrics
    indicators = []

    for predict_key in predict_data.keys():
        if len(predict_data[predict_key]) != 0:
            # Concatenate data
            dft = np.concatenate([arr[predict_key].detach().numpy().flatten() for arr in valid_data_load])
            predict = np.concatenate([arr.detach().numpy().flatten() for arr in predict_data[predict_key]])
            difference = dft - predict

            # Statistical data
            dft_mean, dft_sigma = np.mean(dft), np.std(dft)
            predict_mean, predict_sigma = np.mean(predict), np.std(predict)

            # Calculate evaluation metrics
            mse = np.mean(difference ** 2)
            rmse = np.sqrt(mse)
            mae = np.mean(np.abs(difference))
            r2 = 1 - (np.sum(difference ** 2) / np.sum((dft - dft_mean) ** 2))

            # Output Quality Of Data
            logger.info(
                f"| {predict_key:<16} {dft_mean:>17.4e} {dft_sigma:>17.4f} {predict_mean:>17.4e} {predict_sigma:>17.4f}     |")

            # Save metrics data for later use
            indicators.append((predict_key, rmse, mse, mae, r2))

            # Plot scatter plot (DFT vs Predict)
            if plot:
                from scipy.stats import gaussian_kde
                import matplotlib.pyplot as plt
                plt.rcParams.update({'font.size': 18, 'font.family': 'serif', 'font.serif': ['Times New Roman']})
                plt.rcParams['mathtext.default'] = 'regular'
                plt.figure(figsize=(10, 8))  # Increase image size
                plt.scatter(dft, predict, s=20, label='data')  # Enhance scatter plot effect

                # Diagonal line (y = x, ideal fit line)
                data_min = min(dft.min(), predict.min())
                data_max = max(dft.max(), predict.max())
                plt.plot([data_min, data_max], [data_min, data_max], color='black', linestyle='--', linewidth=1.5,
                         label='$y=x$')

                # Annotate RMSE and R² values
                plt.text(0.05, 0.95, f"RMSE: {rmse:4.4f}{unit[predict_key]}\nR²: {r2:4.4f}", transform=plt.gca().transAxes,
                         fontsize=18, verticalalignment='top', bbox=dict(facecolor='white', alpha=0.8))

                # Set title, axis labels, and grid
                plt.title(f'{predict_key}: dft vs predict', fontsize=22)
                plt.xlabel(f'{predict_key}$_{{dft}}$', fontsize=20)
                plt.ylabel(f'{predict_key}$_{{predict}}$', fontsize=20)
                plt.xlim(data_min, data_max)
                plt.ylim(data_min, data_max)
                plt.tick_params(axis='both', labelsize=18)
                plt.grid(True, linestyle='--', alpha=0.5)
                plt.legend()

                # Save scatter plot
                plt.savefig(f'{predict_key}_scatter.png', dpi=300, bbox_inches='tight')
                plt.close()

                # Plot error distribution (Difference Histogram + KDE)
                plt.figure(figsize=(10, 8))

                # Histogram part
                counts, bins, _ = plt.hist(difference, bins=40, density=True, alpha=0.7, color='#1f77b4',
                                           edgecolor='black')

                # KDE (Kernel Density Estimation) smoothed density curve
                kde = gaussian_kde(difference)
                kde_x = np.linspace(bins.min(), bins.max(), 300)
                plt.plot(kde_x, kde(kde_x), color='red', linewidth=2, label='Density')

                # Annotate RMSE and R² values
                plt.text(0.65, 0.95, f"RMSE: {rmse:4.4f}{unit[predict_key]}\nR²: {r2:4.4f}", transform=plt.gca().transAxes,
                         fontsize=18, verticalalignment='top', bbox=dict(facecolor='white', alpha=0.8))

                # Set title, axis labels, and grid
                plt.title(f'{predict_key} error distribution', fontsize=22)
                plt.xlabel('error (dft - predict)', fontsize=20)
                plt.ylabel('density', fontsize=20)
                plt.tick_params(axis='both', labelsize=18)
                plt.grid(True, linestyle='--', alpha=0.5)
                plt.legend()

                # Save error distribution plot
                plt.savefig(f'ErrorDistributionOf_{predict_key}.png', dpi=300, bbox_inches='tight')
                plt.close()

                # Save data (DFT and Predict)
                np.savetxt(f'{predict_key}.csv', np.column_stack((dft, predict)), header='DFT,Predict', delimiter=',')

    logger.info(f"+----------------------------------------------------------------------------------------------+")

    # Print the Evaluation Indicators header
    logger.info(f"|{'Evaluation Indicators'.center(94)}|")
    logger.info(f"| {'Key':<16} {'RMSE':>17} {'MSE':>17} {'MAE':>17} {'R²':>17}     |")
    logger.info(f"+----------------------------------------------------------------------------------------------+")

    # Output metrics data
    for predict_key, rmse, mse, mae, r2 in indicators:
        logger.info(f"| {predict_key:<16} {rmse:>17.4e} {mse:>17.4e} {mae:>17.4e} {r2:>17.4f}     |")

    logger.info(f"+----------------------------------------------------------------------------------------------+")
